from aptsources.sourceslist import (SourceEntry, Deb822SourceEntry)

def copy_source_entry(orig):
    """ Return a shallow copy of the source entry. """
    if isinstance(orig, Deb822SourceEntry):
        return Deb822SourceEntry(str(orig), file=orig.file)

    return SourceEntry(str(orig), file=orig.file)

def replace_source_entry(orig, **kwargs):
    """Return a copy of the given source entry with replaced field(s)."""
    entry = copy_source_entry(orig)
    for (k, v) in kwargs.items():
        setattr(entry, k, v)
    return entry

def get_source_entry_pocket(source_entry):
    """
    Return the pocket, or if unset return 'release'.

    This always returns the pocket in lowercase.
    """
    parts = source_entry.dist.partition('-')
    return (parts[2] or 'release').lower()

def get_source_entry_suite(source_entry):
    """
    Return the suite, without pocket.

    This always returns the suite in lowercase.
    """
    return source_entry.dist.partition('-')[0]

def get_source_entry_from_list(entries, entry):
    """
    Return the source entry from entries that matches entry, if found.
    Otherwise return None.

    This function uses a modified equality check, i.e. it considers components
    as a set rather than a list so that different ordering does not affect the
    comparison.
    """
    target = replace_source_entry(entry, comps=set(entry.comps))

    for e in entries:
        if replace_source_entry(e, comps=set(e.comps)) == target:
            return e

    return None

def deb822_source_entry_contains(a, b):
    """
    Return True if the source defined by b is already satisfied
    by the source defined by a.

    For example, if source a is:

    Types: deb
    URIs: http://archive.ubuntu.com/ubuntu
    Suites: jammy
    Components: main universe

    and source b is:

    Types: deb
    URIs: http://archive.ubuntu.com/ubuntu
    Suites: jammy
    Components: universe

    Then source a contains source b.

    But if source b was:

    Types: deb-src
    URIs: http://archive.ubuntu.com/ubuntu
    Suites: jammy
    Components: universe

    Then a does not contain b because it does not include deb-src.
    """
    if a.disabled != b.disabled:
        return False

    for attr in ['types', 'comps', 'suites', 'uris']:
        if set(getattr(b, attr)) - set(getattr(a, attr)):
            return False

    return True



#
#  Copyright (c) 2021 Canonical Ltd.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import os
from gettext import gettext as _
import gi
gi.require_version("Gtk", "4.0")
from gi.repository import Gtk,GLib,Gio
from softwareproperties.gtk.utils import setup_ui
from uaclient.api.u.pro.attach.magic.initiate.v1 import initiate
from uaclient.api.u.pro.attach.magic.wait.v1 import MagicAttachWaitOptions, wait
from uaclient.exceptions import MagicAttachTokenError

class DialogUaAttach:
    def __init__(self, parent, datadir, ua_object):
        """setup up the gtk dialog"""
        setup_ui(self, os.path.join(datadir, "gtkbuilder", "dialog-ua-attach.ui"), domain="software-properties")

        self.ua_object = ua_object
        self.dialog = self.dialog_ua_attach
        self.dialog.set_transient_for(parent)

        self.contract_token = None
        self.attaching = False
        self.poll = None
        self.pin = ""

        self.net_monitor = Gio.network_monitor_get_default()
        self.net_monitor.connect("network-changed", self.net_status_changed, 0)
        self.net_status_changed(
            self.net_monitor, self.net_monitor.get_network_available(), 1
        )

        self.magic_radio.set_active(True)
        self.token_radio.set_active(False)
        # See 3to4 migration docs: Adapt to button class hierarchy changes
        self.token_radio.set_group(self.magic_radio)

        self.cancel.connect("clicked", self.on_cancel_clicked)
        self.confirm.connect("clicked", self.on_confirm_clicked)
        self.magic_radio.connect("toggled", self.on_magic_radio_toggled)
        self.token_radio.connect("toggled", self.on_radio_toggled)
        self.token_field.connect("activate", self.on_token_entry_activate)
        self.token_field.connect("changed", self.on_token_typing)

        self.token_status.set_accessible_role(Gtk.AccessibleRole.STATUS)
        self.pin_status.set_accessible_role(Gtk.AccessibleRole.STATUS)

    def present(self):
        self.dialog.present()

    def update_state(self, case = None):
        """
        fail   : called by the attachment callback, and it failed.
        success: called by the attachment callback, and it succeeded.
        expired: called by the token polling when the token expires.
        """
        if self.token_radio.get_active():
            self.confirm.set_sensitive(self.token_field.get_text() != "" and
                                       not self.attaching)
            icon = self.token_status_icon
            spinner = self.token_spinner
            status = self.token_status
        else:
            self.pin_label.set_text(self.pin)
            self.confirm.set_sensitive(self.contract_token != None and
                                       not self.attaching)
            icon = self.pin_status_icon
            spinner = self.pin_spinner
            status = self.pin_status

        if self.attaching:
            spinner.start()
        else:
            spinner.stop()

        def lock_radio_buttons(boolean):
            self.token_radio.set_sensitive(not boolean)
            self.magic_radio.set_sensitive(not boolean)

        lock_radio_buttons(self.attaching)
        self.token_field.set_sensitive(not self.attaching
                                       and self.token_radio.get_active())

        # Unconditionally hide the "other radio section" icon/status.
        # Show icon/status of the "current radio section" only if case is set.
        self.token_status_icon.set_visible(False)
        self.token_status.set_visible(False)
        self.pin_status_icon.set_visible(False)
        self.pin_status.set_visible(False)
        if (case != None):
            icon.set_visible(True)
            status.set_visible(True)

        if (case == "fail"):
            status.set_markup('<span foreground="red">%s</span>' % _('Invalid token'))
            icon.set_from_icon_name('emblem-unreadable')
        elif (case == "success"):
            self.finish()
        elif (case == "pin_validated"):
            status.set_markup('<span foreground="green">%s</span>' % _('Valid token'))
            icon.set_from_icon_name('emblem-default')
            lock_radio_buttons(True)
        elif (case == "expired"):
            status.set_markup(_('Code expired'))
            icon.set_from_icon_name('gtk-dialog-warning')

    def attach(self):
        if self.attaching:
            return

        if self.token_radio.get_active():
            token = self.token_field.get_text()
        else:
            token = self.contract_token

        self.attaching = True
        def on_reply():
            self.attaching = False
            self.update_state("success")
        def on_error(error):
            self.attaching = False
            if self.magic_radio.get_active():
                self.contract_token = None
            self.update_state("fail")
        self.ua_object.Attach(token, reply_handler=on_reply, error_handler=on_error, dbus_interface='com.canonical.UbuntuAdvantage.Manager', timeout=600)
        self.update_state()

    def on_token_typing(self, entry):
        self.confirm.set_sensitive(self.token_field.get_text() != '')

    def on_token_entry_activate(self, entry):
        token = self.token_field.get_text()
        if token != '':
            self.attach()

    def on_confirm_clicked(self, button):
        self.attach()

    def on_cancel_clicked(self, button):
        if self.poll:
            GLib.Thread.unref(self.poll)
        self.dialog.hide()

    def poll_for_magic_token(self):
        options = MagicAttachWaitOptions(magic_token=self.req_id)
        try:
            response = wait(options)
            self.contract_token = response.contract_token
            if self.magic_radio.get_active():
                GLib.idle_add(self.update_state, "pin_validated")
        except MagicAttachTokenError:
            if self.magic_radio.get_active():
                GLib.idle_add(self.update_state, "expired")
        except Exception as e:
            print("Error getting the Ubuntu Pro token: ", e, flush = True)
        finally:
            self.poll = None

    def start_magic_attach(self):
        # Already polling, don't bother the server with a new request.
        if self.poll != None or self.contract_token != None:
            return

        # Request a magic attachment and parse relevants fields from response.
        #  userCode:  The pin the user has to type in <ubuntu.com/pro/attach>;
        #  token:     Identifies the request (used for polling for it).
        try:
            response = initiate()
            self.pin = response.user_code
            self.req_id = response.token
        except Exception as e:
            print("Error retrieving magic token: ", e)
            return
        self.update_state()
        self.poll = GLib.Thread.new("poll", self.poll_for_magic_token)

    def on_radio_toggled(self, button):
        if self.magic_radio.get_active() and self.contract_token:
            self.update_state("pin_validated")
        else:
            self.update_state()

    def on_magic_radio_toggled(self, button):
        self.start_magic_attach()

    # Do not control the radio buttons and confirm button widgets directly,
    # since those former are controlled by update_state and this function must
    # be logically independent of it. Control the net_control_box'es instead.
    def net_status_changed(self, monitor, available, first_run):
        self.no_connection.set_visible(not available)
        self.radio_net_control_box.set_sensitive(available)
        self.confirm_net_control_box.set_sensitive(available)
        if available:
            if self.pin == "":
                self.start_magic_attach()
            elif self.poll == None:
                # wait() timed out without internet; Restart polling.
                self.poll = GLib.Thread.new("poll", self.poll_for_magic_token)

    def finish(self):
        self.dialog.response(Gtk.ResponseType.OK)
        self.dialog.hide()



#  Copyright (c) 2024 Simon Quigley <tsimonq2@ubuntu.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from PyQt6.QtWidgets import (QDialog, QLineEdit, QPushButton, QVBoxLayout, QHBoxLayout, QLabel, QGridLayout, QWidget)
from PyQt6 import QtWidgets, uic
from aptsources.sourceslist import Deb822SourceEntry

class CustomSourceEntryField(QWidget):
    def __init__(self, dialog, key=None, value=None):
        super().__init__()

        self.dialog = dialog
        layout = QGridLayout(self)

        self.entry_key = QLineEdit(self)
        self.entry_key.setPlaceholderText("Key")
        self.entry_val = QLineEdit(self)
        self.entry_val.setPlaceholderText("Value")
        self.button_delete = QPushButton("Delete", self)

        if key is not None:
            self.entry_key.setText(key)

        if value is not None:
            self.entry_val.setText(value)

        self.entry_key.setCursorPosition(0)
        self.entry_val.setCursorPosition(0)

        layout.addWidget(self.entry_key, 0, 0)
        layout.addWidget(self.entry_val, 0, 1)
        layout.addWidget(self.button_delete, 0, 2)

        self.button_delete.clicked.connect(self.delete_field)

    def delete_field(self):
        self.setParent(None)
        self.deleteLater()

    def get_key(self):
        return self.entry_key.text()

    def get_val(self):
        return self.entry_val.text()


class DialogEditDeb822(QDialog):
    def __init__(self, parent, sourceslist, source_entry, datadir):
        super().__init__(parent)

        uic.loadUi("%s/designer/dialog_editdeb822.ui" % datadir, self)

        self.sourceslist = sourceslist
        self.source_entry = source_entry
        self.new_source_entry = None
        self.entry_types = self.findChild(QLineEdit, "entry_types")
        self.entry_uris = self.findChild(QLineEdit, "entry_uris")
        self.entry_suites = self.findChild(QLineEdit, "entry_suites")
        self.entry_comps = self.findChild(QLineEdit, "entry_comps")
        self.entry_comment = self.findChild(QLineEdit, "entry_comment")
        self.button_add_field = self.findChild(QPushButton, "add_button")
        self.button_box = self.findChild(QtWidgets.QDialogButtonBox, "buttonBox")

        self.additional_fields_layout = self.findChild(QGridLayout, "additional_sources_layout")
        self.populate_with_existing_source_entry()

        self.button_add_field.clicked.connect(self.create_custom_source_entry)

    def populate_with_existing_source_entry(self):
        self.entry_types.setText(" ".join(self.source_entry.types))
        self.entry_types.setCursorPosition(0)
        self.entry_uris.setText(" ".join(self.source_entry.uris))
        self.entry_uris.setCursorPosition(0)
        self.entry_suites.setText(" ".join(self.source_entry.suites))
        self.entry_suites.setCursorPosition(0)
        self.entry_comps.setText(" ".join(self.source_entry.comps))
        self.entry_comps.setCursorPosition(0)
        self.entry_comment.setText(self.source_entry.comment.rstrip())
        self.entry_comment.setCursorPosition(0)

        for key, value in self.source_entry.section.tags.items():
            if key not in ["Types", "URIs", "Suites", "Components"]:
                self.append_custom_source_entry_field(key=key, value=value)

    def append_custom_source_entry_field(self, key=None, value=None):
        field = CustomSourceEntryField(self, key, value)
        self.additional_fields_layout.addWidget(field)

    def create_custom_source_entry(self):
        self.append_custom_source_entry_field()

    def get_custom_source_entry_fields(self):
        return [self.additional_fields_layout.itemAt(i).widget() for i in range(self.additional_fields_layout.count())]

    def is_current_input_valid(self):
        if "" in (self.entry_types.text(), self.entry_uris.text(), self.entry_suites.text(), self.entry_comps.text()):
            return False

        for field in self.get_custom_source_entry_fields():
            if "" in (field.get_key(), field.get_val()):
                return False
        return True

    def source_entry_from_input(self):
        source_entry = Deb822SourceEntry(None, self.source_entry.file)

        source_entry.types = self.entry_types.text().split()
        source_entry.uris = self.entry_uris.text().split()
        source_entry.suites = self.entry_suites.text().split()
        source_entry.comps = self.entry_comps.text().split()
        source_entry.comment = self.entry_comment.text()

        for field in self.get_custom_source_entry_fields():
            source_entry.section[field.get_key()] = field.get_val()

        return source_entry

    def run(self):
        res = self.exec()
        if res == QtWidgets.QDialog.DialogCode.Accepted:
            self.new_source_entry = self.source_entry_from_input()

        self.hide()
        return res



#  software-properties cloud-archive support
#
#  Copyright (c) 2013 Canonical Ltd.
#
#  Author: Scott Moser <smoser@ubuntu.org>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import os
import subprocess

from gettext import gettext as _

from softwareproperties.shortcuthandler import (ShortcutHandler, ShortcutException,
                                                InvalidShortcutException)
from softwareproperties.sourceslist import SourcesListShortcutHandler
from softwareproperties.uri import URIShortcutHandler

from urllib.parse import urlparse


RELEASE_MAP = {
    'folsom': 'precise',
    'grizzly': 'precise',
    'havana': 'precise',
    'icehouse': 'precise',
    'juno': 'trusty',
    'kilo': 'trusty',
    'liberty': 'trusty',
    'mitaka': 'trusty',
    'newton': 'xenial',
    'ocata': 'xenial',
    'pike': 'xenial',
    'queens': 'xenial',
    'rocky': 'bionic',
    'stein': 'bionic',
    'train': 'bionic',
    'ussuri': 'bionic',
    'victoria': 'focal',
    'wallaby': 'focal',
    'xena': 'focal',
    'yoga': 'focal',
    'zed': 'jammy',
    'antelope': 'jammy',
    'bobcat': 'jammy',
    'caracal': 'jammy',
    'dalmatian': 'noble',
    'epoxy': 'noble',
    'flamingo': 'noble',
    'gazpacho': 'noble',
}
UCA = "Ubuntu Cloud Archive"
WEB_LINK = 'https://wiki.ubuntu.com/OpenStack/CloudArchive'

UCA_ARCHIVE = "http://ubuntu-cloud.archive.canonical.com/ubuntu"
UCA_PREFIXES = ['cloud-archive', 'uca']
UCA_VALID_POCKETS = ['updates', 'proposed']
UCA_DEFAULT_POCKET = UCA_VALID_POCKETS[0]

UCA_KEYRING_PACKAGE = 'ubuntu-cloud-keyring'


class CloudArchiveShortcutHandler(ShortcutHandler):
    def __init__(self, shortcut, **kwargs):
        super(CloudArchiveShortcutHandler, self).__init__(shortcut, **kwargs)
        self.caname = None

        # one of these will set caname and pocket, and maybe _source_entry
        if not any((self._match_uca(shortcut),
                    self._match_uri(shortcut),
                    self._match_sourceslist(shortcut))):
            msg = (_("not a valid cloud-archive format: '%s'") % shortcut)
            raise InvalidShortcutException(msg)

        self.caname = self.caname.lower()

        self._filebase = "cloudarchive-%s" % self.caname

        self.pocket = self.pocket.lower()
        if not self.pocket in UCA_VALID_POCKETS:
            msg = (_("not a valid cloud-archive pocket: '%s'") % self.pocket)
            raise ShortcutException(msg)

        if not self.caname in RELEASE_MAP:
            msg = (_("not a valid cloud-archive: '%s'") % self.caname)
            raise ShortcutException(msg)

        codename = RELEASE_MAP[self.caname]
        validnames = set((codename, os.getenv("CA_ALLOW_CODENAME") or codename))
        if self.codename not in validnames:
            msg = (_("cloud-archive for %s only supported on %s") %
                   (self.caname.capitalize(), codename.capitalize()))
            raise ShortcutException(msg)

        self._description = f'{UCA} for OpenStack {self.caname.capitalize()}'
        if self.pocket == 'proposed':
            self._description += ' [proposed]'

        if not self._source_entry:
            dist = f'{self.codename}-{self.pocket}/{self.caname}'
            comps = ' '.join(self.components) or 'main'
            line = f'{self.binary_type} {UCA_ARCHIVE} {dist} {comps}'
            self._set_source_entry(line)

    @property
    def description(self):
        return self._description

    @property
    def web_link(self):
        return WEB_LINK

    def add_key(self):
        # UCA provides its repo keys in a package
        subprocess.run(f'apt-get install -y {UCA_KEYRING_PACKAGE}'.split(), check=True)

    def _encode_filebase(self, suffix=None):
        # ignore suffix
        return super(CloudArchiveShortcutHandler, self)._encode_filebase()

    def _match_uca(self, shortcut):
        (prefix, _, uca) = shortcut.rpartition(':')
        if not prefix.lower() in UCA_PREFIXES:
            return False

        (caname, _, pocket) = uca.partition('-')
        if not caname:
            return False

        self.caname = caname
        self.pocket = pocket or self.pocket or UCA_DEFAULT_POCKET
        return True

    def _match_uri(self, shortcut):
        try:
            return self._match_handler(URIShortcutHandler(shortcut))
        except InvalidShortcutException:
            return False

    def _match_sourceslist(self, shortcut):
        try:
            return self._match_handler(SourcesListShortcutHandler(shortcut))
        except InvalidShortcutException:
            return False

    def _match_handler(self, handler):
        parsed = urlparse(handler.SourceEntry().uri)
        if parsed.hostname != urlparse(UCA_ARCHIVE).hostname:
            return False

        (codename, _, caname) = handler.SourceEntry().dist.partition('/')
        (codename, _, pocket) = codename.partition('-')

        if not all((codename, caname)):
            return False

        self.caname = caname
        self.pocket = pocket or self.pocket or UCA_DEFAULT_POCKET

        self._set_source_entry(handler.SourceEntry().line)
        return True



# dialog_apt_key.py.in - edit the apt keys
#
#  Copyright (c) 2004 Canonical
#
#  Author: Michael Vogt <mvo@debian.org>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import gettext
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile

from subprocess import PIPE

# gettext convenient
_ = gettext.gettext
N_ = lambda e: e

# some known keys
N_("Ubuntu Archive Automatic Signing Key <ftpmaster@ubuntu.com>")
N_("Ubuntu CD Image Automatic Signing Key <cdimage@ubuntu.com>")
N_("Ubuntu Archive Automatic Signing Key (2012) <ftpmaster@ubuntu.com>")
N_("Ubuntu CD Image Automatic Signing Key (2012) <cdimage@ubuntu.com>")
N_("Ubuntu Extras Archive Automatic Signing Key <ftpmaster@ubuntu.com>")
PARTS=pathlib.Path("etc/apt/trusted.gpg.d")

class AptAuth:

    def __init__(self, rootdir="/"):
        self.rootdir = pathlib.Path(rootdir)

    def run_gpg(self, *args):
        return subprocess.run(["gpg",
                               "--homedir", self.home,
                               "--no-options",
                               "--no-default-keyring",
                               "--no-auto-check-trustdb",
                               "--trust-model", "always",
                               "--with-colons",
                               "--batch",
                               "--fixed-list-mode",
                               ] + list(args), stdout=PIPE, text=True)

    def list(self):
        res = []
        for filename in os.listdir(self.rootdir / PARTS):
            if not any(filename.endswith(x) for x in (".asc", ".gpg")):
                continue
            filepath = self.rootdir / PARTS / filename
            try:
                process = subprocess.run(["sq", "--cert-store=none", "--key-store=none",
                                          "keyring", "list", str(filepath)], stdout=PIPE, text=True)
                res.append("%s\n  %s" % (filepath, "\n  ".join(process.stdout.strip().split("\n"))))
            except OSError:
                # We can only hope and pray gpg spawns no daemons that take over the home dir
                with tempfile.TemporaryDirectory() as self.home:
                    keyring = filepath
                    if keyring.suffix == ".asc":
                        keyring = pathlib.Path(self.home) / "dearmored-key.gpg"
                        self.run_gpg("-o", keyring, "--dearmor", filepath)
                    process = self.run_gpg("--keyring", keyring, "--list-keys")
                lines = []
                for line in process.stdout.splitlines():
                    fields = line.split(":")
                    if fields[0] == "fpr":
                        fpr = fields[9]
                    if fields[0] == "uid":
                        uid = fields[9]

                        lines.append(f"{fpr} {uid}")
                res.append("%s\n  %s" % (filepath, "\n  ".join(lines)))

        return res

    def add(self, filename):
        out = self.rootdir / PARTS / os.path.basename(filename)
        print("Add", filename, out, file=sys.stderr)
        if not os.path.exists(out):
            print("Copy", filename, out, file=sys.stderr)
            try:
                shutil.copyfile(filename, out)
            except Exception as e:
                print(e)
            return True
        return False

    def rm(self, key):
        print("Removing", key)
        try:
            os.unlink(key)
        except OSError:
            return True



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from gi.repository import GLib

import apt_pkg
import aptsources.distro
import aptsources.sourceslist

import dbus
import logging
import glob
import os
import subprocess
import sys
import threading
import time
import unittest
import pathlib

from dbus.mainloop.glib import DBusGMainLoop

sys.path.insert(0, "../")
from softwareproperties.dbus.SoftwarePropertiesDBus import (
    SoftwarePropertiesDBus, DBUS_BUS_NAME, DBUS_PATH, DBUS_INTERFACE_NAME)
from softwareproperties import (
    UPDATE_INST_SEC, UPDATE_DOWNLOAD, UPDATE_NOTIFY)

try:
    DPKG_ARCH = subprocess.check_output(
        ["dpkg", "--print-architecture"]).strip().decode("utf-8")
except subprocess.CalledProcessError:
    sys.stderr.write("WARNING: Failed to read dpkg arch")
    DPKG_ARCH = None

if DPKG_ARCH in ("i386", "amd64"):
    PRIMARY_MIRROR = "http://archive.ubuntu.com/ubuntu"
else:
    PRIMARY_MIRROR = "http://ports.ubuntu.com/ubuntu-ports"


def get_test_source_line():
    distro_release = get_distro_release()
    return "deb %s %s main restricted #"\
           " comment with unicode äöü" % (PRIMARY_MIRROR, distro_release)


def get_dpkg_arch():
    return subprocess.check_output(
        ["dpkg", "--print-architecture"]).strip().decode("utf-8")


def get_distro_release():
    return "bionic"


def clear_apt_config():
    etc_apt = os.path.join(os.path.dirname(__file__), "aptroot", "etc", "apt")
    for dirpath, dirnames, filenames in os.walk(etc_apt):
        for name in filenames:
            path = os.path.join(dirpath, name)
            if os.path.isfile(path):
                os.unlink(path)

    for d in ["apt.conf.d", "sources.list.d", "trusted.gpg.d", "auth.conf.d"]:
        os.makedirs(os.path.join(etc_apt, d), exist_ok=True)

def create_sources_list():
    s = get_test_source_line() + "\n"
    name = os.path.join(os.path.dirname(__file__),
                        "aptroot", "etc", "apt", "sources.list")
    dirname = os.path.dirname(name)
    if not os.path.exists(dirname):
        pathlib.Path(dirname).mkdir(parents=True)
    with open(name, "w") as f:
        f.write(s)
    return name


def session_bus_thread():
    DBusGMainLoop(set_as_default=True)
    loop = GLib.MainLoop()
    threading.current_thread().loop = loop

    bus = dbus.SessionBus(private=True)
    bus.set_exit_on_disconnect(False)

    rootdir = os.path.join(os.path.dirname(__file__), "aptroot")
    spd = SoftwarePropertiesDBus(bus, rootdir=rootdir)
    spd.enforce_polkit = False
    loop.run()
    bus.close()


class TestDBus(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for k in apt_pkg.config.keys():
            apt_pkg.config.clear(k)
        apt_pkg.init()
        clear_apt_config()
        # create sources.list file
        create_sources_list()
        cls.get_distro = aptsources.distro.get_distro
        aptsources.distro.get_distro = lambda *a, **b: aptsources.distro.UbuntuDistribution(id="Ubuntu", codename="bionic", description="18.04", release="bionic")
        cls.thread = threading.Thread(target=session_bus_thread)
        cls.thread.loop = None
        cls.thread.start()
        while not cls.thread.loop or not cls.thread.loop.is_running():
            time.sleep(1)

    @classmethod
    def tearDownClass(cls):
        cls.thread.loop.quit()
        cls.thread.join()
        aptsources.distro.get_distro = cls.get_distro

    def setUp(self):
        # keep track of signal emissions
        self.sources_list_count = 0
        self.distro_release = get_distro_release()
        create_sources_list()
        self._sourceslist = aptsources.sourceslist.SourcesList()
        # create the client proxy
        self.bus = dbus.SessionBus(private=True, mainloop=DBusGMainLoop())
        self.bus.set_exit_on_disconnect(False)
        self.iface = dbus.Interface(self.bus.get_object(DBUS_BUS_NAME, DBUS_PATH),
                                    DBUS_INTERFACE_NAME)
        self._signal_id = self.iface.connect_to_signal(
            "SourcesListModified", self._on_sources_list_modified)

    def tearDown(self):
        # ensure we remove the "modified" signal again
        self._signal_id.remove()
        self.bus.close()

    def _on_sources_list_modified(self):
        #print("_on_modified_sources_list")
        self.sources_list_count += 1

    @property
    def sourceslist(self):
        self._sourceslist.refresh()
        return ''.join([str(e) for e in self._sourceslist])

    @property
    def enabled_sourceslist(self):
        self._sourceslist.refresh()
        return ''.join([str(e) for e in self._sourceslist
                        if not e.invalid and not e.disabled])

    def _debug_sourceslist(self, text=""):
        logging.debug("sourceslist: %s '%s'" % (text, self.sourceslist))

    # this is an async call - give it a few seconds to catch up with what we expect
    def _assert_eventually(self, prop, n):
        for i in range(9):
            if getattr(self, prop) == n:
                self.assertEqual(getattr(self, prop), n)
            else:
                time.sleep(1)
        # nope, you die now
        self.assertEqual(getattr(self, prop), n)

    def test_enable_disable_component(self):
        # ensure its not there
        self.assertNotIn("universe", self.sourceslist)
        # enable
        self.iface.EnableComponent("universe")
        self._debug_sourceslist("2")
        self.assertIn("universe", self.sourceslist)
        # disable again
        self.iface.DisableComponent("universe")
        self._debug_sourceslist("3")
        self.assertNotIn("universe", self.sourceslist)
        self._assert_eventually("sources_list_count", 2)

    def test_enable_enable_disable_source_code_sources(self):
        # ensure its not there
        self._debug_sourceslist("4")
        self.assertNotIn('deb-src', self.enabled_sourceslist)
        # enable
        self.iface.EnableSourceCodeSources()
        self._debug_sourceslist("5")
        self.assertIn('deb-src', self.enabled_sourceslist)
        # disable again
        self.iface.DisableSourceCodeSources()
        self._debug_sourceslist("6")
        self.assertNotIn('deb-src', self.enabled_sourceslist)
        self._assert_eventually("sources_list_count", 3)

    def test_enable_child_source(self):
        child_source = "%s-updates" % self.distro_release
        # ensure its not there
        self._debug_sourceslist("7")
        self.assertNotIn(child_source, self.sourceslist)
        # enable
        self.iface.EnableChildSource(child_source)
        self._debug_sourceslist("8")
        self.assertIn(child_source, self.sourceslist)
        # disable again
        self.iface.DisableChildSource(child_source)
        self._debug_sourceslist("9")
        self.assertNotIn(child_source, self.sourceslist)
        self._assert_eventually("sources_list_count", 2)

    def test_toggle_source(self):
        # test toggle
        source = get_test_source_line()
        self.iface.ToggleSourceUse(source)
        self._debug_sourceslist("10")
        primary_debline = "# deb %s" % PRIMARY_MIRROR
        self.assertIn(primary_debline, self.sourceslist)
        # to disable the line again, we need to match the new "#"
        source = "# " + source
        self.iface.ToggleSourceUse(source)
        self._debug_sourceslist("11")
        self.assertNotIn(primary_debline, self.sourceslist)

        self._assert_eventually("sources_list_count", 2)

    def test_replace_entry(self):
        # test toggle
        source = get_test_source_line()
        source_new = "deb http://xxx/ %s" % self.distro_release
        self.iface.ReplaceSourceEntry(source, source_new)
        self._debug_sourceslist("11")
        self.assertIn(source_new, self.sourceslist)
        self.assertNotIn(source, self.sourceslist)
        self._assert_eventually("sources_list_count", 1)
        self.iface.ReplaceSourceEntry(source_new, source)
        self._assert_eventually("sources_list_count", 2)

    def test_popcon(self):
        # ensure its set to no
        popcon_p = os.path.join(os.path.dirname(__file__),
                              "aptroot", "etc", "popularity-contest.conf")
        with open(popcon_p) as f:
            popcon = f.read()
            self.assertIn('PARTICIPATE="no"', popcon)
        # toggle
        self.iface.SetPopconPariticipation(True)
        with open(popcon_p) as f:
            popcon = f.read()
            self.assertIn('PARTICIPATE="yes"', popcon)
            self.assertNotIn('PARTICIPATE="no"', popcon)
        # and back
        self.iface.SetPopconPariticipation(False)
        with open(popcon_p) as f:
            popcon = f.read()
            self.assertNotIn('PARTICIPATE="yes"', popcon)
            self.assertIn('PARTICIPATE="no"', popcon)

    def test_updates_automation(self):
        states = [UPDATE_INST_SEC, UPDATE_DOWNLOAD, UPDATE_NOTIFY]
        # security
        self.iface.SetUpdateAutomationLevel(states[0])
        cfg = os.path.join(os.path.dirname(__file__),
                           "aptroot", "etc", "apt", "apt.conf.d",
                           "10periodic")
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Unattended-Upgrade "1";', config)
        # download
        self.iface.SetUpdateAutomationLevel(states[1])
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Unattended-Upgrade "0";', config)
            self.assertIn('APT::Periodic::Download-Upgradeable-Packages "1";', config)
        # notify
        self.iface.SetUpdateAutomationLevel(states[2])
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Unattended-Upgrade "0";', config)
            self.assertIn('APT::Periodic::Download-Upgradeable-Packages "0";', config)

    def test_updates_interval(self):
        # interval
        self.iface.SetUpdateInterval(0)
        cfg = os.path.join(os.path.dirname(__file__),
                           "aptroot", "etc", "apt", "apt.conf.d",
                           "10periodic")
        with open(cfg) as f:
            config = f.read()
            self.assertTrue(
                'APT::Periodic::Update-Package-Lists' not in config or
                'APT::Periodic::Update-Package-Lists "0";' in config)
        self.iface.SetUpdateInterval(1)
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Update-Package-Lists "1";', config)
        self.iface.SetUpdateInterval(0)
        with open(cfg) as f:
            config = f.read()
            self.assertIn('APT::Periodic::Update-Package-Lists "0";', config)

    def test_add_remove_source_by_line(self):
        # add invalid
        res = self.iface.AddSourceFromLine("xxx")
        self.assertFalse(res)
        # add real
        s = "deb https://ppa.launchpadcontent.net/ foo bar"
        self.iface.AddSourceFromLine(s)
        self.assertIn(s, self.sourceslist)
        self.assertIn(s.replace("deb", "# deb-src"), self.sourceslist)
        # remove again
        self.iface.RemoveSource(s)
        self.iface.RemoveSource(s.replace("deb", "deb-src"))
        self.assertNotIn(s, self.sourceslist)
        self.assertNotIn(s.replace("deb", "# deb-src"), self.sourceslist)
        self._assert_eventually("sources_list_count", 4)

    def test_add_gpg_key(self):
        # clean
        gpg_glob = os.path.join(os.path.dirname(__file__),
                              "aptroot", "etc", "apt", "trusted.gpg.d", "*.asc")
        trusted_gpg_d = os.path.join(os.path.dirname(__file__),
                                     "aptroot", "etc", "apt", "trusted.gpg.d/")
        testkey = os.path.join(os.path.dirname(__file__),
                               "data", "testkey.asc")
        for f in glob.glob(gpg_glob):
            os.remove(f)
        self.assertTrue(len(glob.glob(gpg_glob)) == 0)
        # add key from file
        res = self.iface.AddKey(os.path.join(os.path.dirname(__file__),
                                             "data", "testkey.asc"))
        self.assertTrue(res)
        self.assertEqual(len(glob.glob(gpg_glob)), 1)
        self.assertNotEqual(os.path.getsize(trusted_gpg_d + "testkey.asc"), 0)
        # remove the key
        res = self.iface.RemoveKey(trusted_gpg_d + "testkey.asc")
        self.assertTrue(res)
        self.assertEqual(len(glob.glob(gpg_glob)), 0)
        # add from data
        with open(testkey) as keyfile:
            data = keyfile.read()
            res = self.iface.AddKeyFromData(data)
            self.assertTrue(res)
            self.assertEqual(len(os.listdir(trusted_gpg_d)), 1)
        # remove the key
        res = self.iface.RemoveKey(trusted_gpg_d + "/46caf96d27ee1eebcf139483275a5c1bb41590474e35f8bb4ae3cceb413f7518.gpg")
        self.assertTrue(res)
        self.assertEqual(len(os.listdir(trusted_gpg_d)), 1)
        # test nonsense
        res = self.iface.AddKeyFromData("nonsens")
        self.assertFalse(res)


if __name__ == "__main__":
    if "-d" in sys.argv:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    unittest.main()



#  software-properties PPA support, using launchpadlib
#
#  Copyright (c) 2019 Canonical Ltd.
#
#  Original Author: Michael Vogt <mvo@debian.org>
#  Rewrite: Dan Streetman <ddstreet@canonical.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from gettext import gettext as _

from launchpadlib.launchpad import Launchpad
from lazr.restfulclient.errors import (NotFound, BadRequest, Unauthorized)

from softwareproperties.shortcuthandler import (ShortcutHandler, ShortcutException,
                                                InvalidShortcutException)
from softwareproperties.sourceslist import SourcesListShortcutHandler
from softwareproperties.uri import URIShortcutHandler

from aptsources.sourceslist import Deb822SourceEntry

from urllib.parse import urlparse


PPA_URI_FORMAT = 'https://ppa.launchpadcontent.net/{team}/{ppa}/ubuntu/'
PRIVATE_PPA_URI_FORMAT = 'https://private-ppa.launchpadcontent.net/{team}/{ppa}/ubuntu/'
PPA_VALID_HOSTNAMES = [
    urlparse(PPA_URI_FORMAT).hostname,
    urlparse(PRIVATE_PPA_URI_FORMAT).hostname,
    # Old hostnames.
    'ppa.launchpad.net',
    'private-ppa.launchpad.net',
]

PPA_VALID_COMPS = ['main', 'main/debug']


class PPAShortcutHandler(ShortcutHandler):
    def __init__(self, shortcut, login=False, lp=None, **kwargs):
        super(PPAShortcutHandler, self).__init__(shortcut, deb822=True, **kwargs)
        self._lp_anon = not login
        self._signing_key_data = None

        self._lp = lp                       # LP object
        self._lpteam = None                 # Person/Team LP object
        self._lpppa = None                  # PPA Archive LP object

        self._is_sourceslist = False

        # one of these will set teamname and ppaname, and maybe _source_entry
        if not self._match_ppa(shortcut):
            # The input is either a sources.list line or a URI. Both cases lead
            # to the SourcesListShortcutHandler being used, so unset
            # self.deb822 (LP: #2037210).
            self.deb822 = False

            if not any((self._match_uri(shortcut),
                        self._match_sourceslist(shortcut))):
                msg = (_("ERROR: '%s' is not a valid ppa format") % shortcut)
                raise InvalidShortcutException(msg)

        self._filebase = "%s-ubuntu-%s" % (self.teamname, self.ppaname)
        self._set_auth()

        # Make sure we can find/access the PPA, lp:#1965180
        if self._is_sourceslist:
            try:
                self.lpppa
            except ShortcutException:
                raise InvalidShortcutException(_("ERROR: Can't find ppa"))

        if not self._source_entry:
            comps = self.components
            if not comps:
                comps = ['main']
                if self.lpppa.publish_debug_symbols:
                    print("PPA publishes dbgsym, you may need to include 'main/debug' component")
                    # comps += ['main/debug']

            uri_format = PRIVATE_PPA_URI_FORMAT if self.lpppa.private else PPA_URI_FORMAT
            uri = uri_format.format(team=self.teamname, ppa=self.ppaname)

            entry = Deb822SourceEntry(None, '')
            entry.types = [self.binary_type]
            entry.uris = [uri]
            entry.suites = [self.dist]
            entry.comps = comps

            self._set_source_entry(str(entry))

    @property
    def lp(self):
        if not self._lp:
            if self._lp_anon:
                login_func = Launchpad.login_anonymously
            else:
                login_func = Launchpad.login_with
            self._lp = login_func("%s.%s" % (self.__module__, self.__class__.__name__),
                                  service_root='production',
                                  version='devel')
        return self._lp

    @property
    def lpteam(self):
        if not self._lpteam:
            try:
                self._lpteam = self.lp.people(self.teamname)
            except NotFound:
                msg = (_("ERROR: user/team '%s' not found (use --login if private)") % self.teamname)
                raise ShortcutException(msg)
            except Unauthorized:
                msg = (_("ERROR: invalid user/team name '%s'") % self.teamname)
                raise ShortcutException(msg)
        return self._lpteam

    @property
    def lpppa(self):
        if not self._lpppa:
            try:
                self._lpppa = self.lpteam.getPPAByName(name=self.ppaname)
            except NotFound:
                msg = (_("ERROR: ppa '%s/%s' not found (use --login if private)") %
                       (self.teamname, self.ppaname))
                raise ShortcutException(msg)
            except BadRequest:
                msg = (_("ERROR: invalid ppa name '%s'") % self.ppaname)
                raise ShortcutException(msg)
        return self._lpppa

    @property
    def description(self):
        return self.lpppa.description

    @property
    def web_link(self):
        return self.lpppa.web_link

    @property
    def trustedparts_content(self):
        if not self._signing_key_data:
            key = self.lpppa.getSigningKeyData()
            fingerprint = self.lpppa.signing_key_fingerprint

            if not fingerprint:
                print(_("Warning: could not get PPA signing_key_fingerprint from LP, using anyway"))
            elif 'redacted' in fingerprint:
                print(_("Private PPA fingerprint redacted, using key anyway (LP: #1879781)"))
            elif not fingerprint in self.fingerprints(key):
                msg = (_("Fingerprints do not match, not importing: '%s' != '%s'") %
                       (fingerprint, ','.join(self.fingerprints(key))))
                raise ShortcutException(msg)

            self._signing_key_data = key
        return self._signing_key_data

    def SourceEntry(self, pkgtype=None):
        entry = super(PPAShortcutHandler, self).SourceEntry(pkgtype=pkgtype)
        if pkgtype != self.source_type or self.components:
            return entry

        # 'main/debug' is needed to get dbgsyms from ppas,
        # but it's not a valid component for ppa deb-src lines.
        # Sigh.
        entry.comps = list(set(entry.comps) - set(['main/debug']))
        return entry

    def _set_source_entry(self, line):
        super(PPAShortcutHandler, self)._set_source_entry(line)

        invalid_comps = set(self.SourceEntry().comps) - set(PPA_VALID_COMPS)
        if invalid_comps:
            print(_("Warning: components '%s' not valid for PPA") % ' '.join(invalid_comps))

    def _match_ppa(self, shortcut):
        (prefix, _, ppa) = shortcut.rpartition(':')
        if not prefix.lower() == 'ppa':
            return False

        (teamname, _, ppaname) = ppa.partition('/')
        teamname = teamname.lstrip('~')
        if '/' in ppaname:
            (ubuntu, _, ppaname) = ppaname.partition('/')
            if ubuntu.lower() != 'ubuntu':
                # PPAs only support ubuntu
                return False
            if '/' in ppaname:
                # Path is too long for valid ppa
                return False

        self.teamname = teamname
        self.ppaname = ppaname or 'ppa'
        return True

    def _match_uri(self, shortcut):
        try:
            return self._match_handler(URIShortcutHandler(shortcut))
        except InvalidShortcutException:
            return False

    def _match_sourceslist(self, shortcut):
        try:
            handler = self._match_handler(SourcesListShortcutHandler(shortcut))
        except InvalidShortcutException:
            return False
        self._is_sourceslist = True
        return handler

    def _match_handler(self, handler):
        parsed = urlparse(handler.SourceEntry().uri)
        if not parsed.hostname in PPA_VALID_HOSTNAMES:
            return False

        path = parsed.path.strip().strip('/').split('/')
        if len(path) < 2:
            return False
        self.teamname = path[0]
        self.ppaname = path[1]

        self._username = handler.username
        self._password = handler.password

        self._set_source_entry(handler.SourceEntry().line)
        return True

    def _set_auth(self):
        if self._lp_anon or not self.lpppa.private:
            return

        if self._username and self._password:
            return

        try:
            # Named ops work against actual person, not the me alias object
            me = self.lp.people(self.lp.me.name)
            url = me.getArchiveSubscriptionURL(archive=self.lpppa)
        except Unauthorized:
            msg = (_("Could not find PPA subscription for ppa:%s/%s, you may need to request access") %
                   (self.teamname, self.ppaname))
            raise ShortcutException(msg)
        else:
            parsed = urlparse(url)
            self._username = parsed.username
            self._password = parsed.password



#  Copyright (c) 2019 Canonical Ltd.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from aptsources.distro import get_distro
from aptsources.sourceslist import SourceEntry

from softwareproperties.sourceslist import SourcesListShortcutHandler


class URIShortcutHandler(SourcesListShortcutHandler):
    def __init__(self, shortcut, **kwargs):
        (uri, _, line_comps) = shortcut.strip().partition(' ')

        line_comps = set(line_comps.split() if line_comps else [])
        param_comps = set(kwargs.get('components', []))
        comps = list(line_comps | param_comps)

        if not comps:
            # if no comps provided, we default to 'main'
            comps = ['main']

        kwargs['components'] = comps

        suite = kwargs.get('codename')
        pocket = kwargs.get('pocket')

        s = SourceEntry('')
        s.invalid = False
        s.uri = uri
        s.type = get_distro().binary_type
        s.dist = suite or get_distro().codename
        if pocket is not None:
            s.dist = '{}-{}'.format(s.dist, pocket)

        super(URIShortcutHandler, self).__init__(str(s), **kwargs)


# vi: ts=4 expandtab



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import apt_pkg
import aptsources

import logging
import os
import shutil
import sys
import tempfile
import unittest

from unittest.mock import patch

sys.path.insert(0, "../")
from softwareproperties.SoftwareProperties import (
    SoftwareProperties)

class SoftwarePropertiesTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for k in apt_pkg.config.keys():
            apt_pkg.config.clear(k)
        apt_pkg.init()
        cls.get_distro = aptsources.distro.get_distro
        aptsources.distro.get_distro = lambda *a, **b: aptsources.distro.UbuntuDistribution(id="Ubuntu", codename="bionic", description="18.04", release="bionic")

    @classmethod
    def tearDownClass(cls):
        aptsources.distro.get_distro = cls.get_distro

    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.sp = SoftwareProperties()
        self.mock_key = os.path.join(self.temp_dir, u"määäp.asc")
        with open(self.mock_key, "wb") as fp:
            fp.write(u"bäää".encode("utf-8"))

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_add_key_str(self):
        with patch.object(self.sp, "apt_key") as mock_apt_key:
            self.sp.add_key(self.mock_key)
        self.assertTrue(mock_apt_key.add.called)

    def test_add_key_bytes(self):
        with patch.object(self.sp, "apt_key") as mock_apt_key:
            mock_name_as_bytes = self.mock_key.encode("utf-8")
            self.sp.add_key(mock_name_as_bytes)
        self.assertTrue(mock_apt_key.add.called)


if __name__ == "__main__":
    if "-d" in sys.argv:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    unittest.main()



#  Copyright (c) 2019 Canonical Ltd.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from gettext import gettext as _

from aptsources.sourceslist import SourceEntry

from softwareproperties.shortcuthandler import (ShortcutHandler, InvalidShortcutException)

from urllib.parse import urlparse


SOURCESLIST_FILE_PREFIX = "archive_uri"

class SourcesListShortcutHandler(ShortcutHandler):
    def __init__(self, shortcut, **kwargs):
        super(SourcesListShortcutHandler, self).__init__(shortcut, **kwargs)

        entry = SourceEntry(shortcut)
        if entry.invalid:
            raise InvalidShortcutException(_("Invalid sources.list line: '%s'") % shortcut)

        uri = entry.uri
        if not self.is_valid_uri(uri):
            raise InvalidShortcutException(_("Invalid URI: '%s'") % uri)

        self.components = list(set(self.components) | set(entry.comps))

        parsed = urlparse(uri)

        self._username = parsed.username
        self._password = parsed.password

        entry.uri = self.uri_strip_auth(entry.uri)
        # must set _filebase first; _set_source_entry uses it to set entry.file
        self._filebase = f"{SOURCESLIST_FILE_PREFIX}-{entry.uri}"
        self._set_source_entry(str(entry))


# vi: ts=4 expandtab



#!/usr/bin/env python

import gi
gi.require_version("Gdk", "4.0")
gi.require_version("Gtk", "4.0")
from gi.repository import GObject, Gio, Gtk
import os
from gettext import gettext as _
import gettext
try:
    from urllib.request import url2pathname
except ImportError:
    from urllib import url2pathname

from softwareproperties.gtk.utils import (
    setup_ui,
)

from aptsources.sourceslist import SourcesList, SourceEntryMatcher


# Holds `SourceItem` markup text for the GTK4's new Gio.ListMode.
class SourceItem(GObject.Object):
    def __init__(self, markup_text):
        super().__init__()
        self.markup_text = markup_text


class DialogAddSourcesList:
    def __init__(self, parent, sourceslist, source_renderer,
                 get_comparable, datadir, file):
        print(file)
        self.parent = parent
        self.source_renderer = source_renderer
        self.sourceslist = sourceslist
        self.get_comparable = get_comparable
        self.file = self.format_uri(file)

        setup_ui(self, os.path.join(datadir, "gtkbuilder", "dialog-add-sources-list.ui"), domain="software-properties")

        self.dialog = self.dialog_add_sources_list
        self.label = self.label_sources
        self.listview = self.listview_sources
        self.scrolled = self.scrolled_window

        if self.parent is not None:
            self.dialog.set_transient_for(parent)
        else:
            self.dialog.set_title(_("Add Software Channels"))

        self.setup_model()

        # Parse the source.list file
        try:
            self.new_sources = SingleSourcesList(self.file)
        except:
            self.error()
            return

        # show the found channels or an error message
        if len(self.new_sources.list) > 0:
            counter = 0

            for source in self.new_sources.list:
                if source.invalid or source.disabled:
                    continue
                self.new_sources.matcher.match(source)
            # sort the list
            self.new_sources.list.sort(key=self.get_comparable)

            for source in self.new_sources.list:
                if source.invalid or source.disabled:
                    continue
                counter += 1
                line = self.source_renderer(source)
                source_item = SourceItem(line)
                self.store.append(source_item)
            if counter == 0:
                self.error()
                return

            header = gettext.ngettext("Install software additionally or "
                                      "only from this source?",
                                      "Install software additionally or "
                                      "only from these sources?",
                                      counter)
            body = _("You can either add the following sources or replace your "
                     "current sources by them. Only install software from "
                     "trusted sources.")
            self.label.set_markup("<big><b>%s</b></big>\n\n%s" % (header, body))
        else:
            self.error()
            return

    def setup_model(self):
        self.store = Gio.ListStore(item_type=SourceItem)
        self.selection_model = Gtk.NoSelection(model=self.store)
        self.listview.set_model(self.selection_model)

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup)
        factory.connect("bind", self._on_factory_bind)
        factory.connect("unbind", self._on_factory_unbind)
        factory.connect("teardown", self._on_factory_teardown)
        self.listview.set_factory(factory)

    def _on_factory_setup(self, factory, list_item):
        label = Gtk.Label()
        label.set_use_markup(True)
        label.set_xalign(0.0)  # Left align
        label.set_margin_start(6)
        label.set_margin_end(6)
        label.set_margin_top(6)
        label.set_margin_bottom(6)
        list_item.set_child(label)

    def _on_factory_bind(self, factory, list_item):
        source_item = list_item.get_item()
        label = list_item.get_child()
        label.set_markup(source_item.markup_text)

    def _on_factory_unbind(self, factory, list_item):
        # Clear the label content when unbinding
        label = list_item.get_child()
        label.set_text("")

    def _on_factory_teardown(self, factory, list_item):
        pass

    def error(self):
        self.button_add.hide()
        self.button_cancel.set_label(_("Close"))
        self.button_replace.hide()
        self.scrolled.hide()
        header = _("There are no sources to install software from")
        body = _("The file '%s' does not contain any valid "
                 "software sources." % self.file)
        self.label.set_markup("%s\n\n<small>%s</small>" % (header, body))

    def present(self, on_response):
        def _on_response(_dlg, response_id):
            self.dialog.destroy()
            on_response(response_id, self.new_sources)
        self.dialog.connect("response", _on_response)
        self.dialog.present()

    def format_uri(self, uri):
        path = url2pathname(uri) # escape special chars
        path = path.strip('\r\n\x00') # remove \r\n and NULL
        if path.startswith('file:\\\\\\'): # windows
            path = path[8:] # 8 is len('file:///')
        elif path.startswith('file://'): #nautilus, rox
            path = path[7:] # 7 is len('file://')
        elif path.startswith('file:'): # xffm
            path = path[5:] # 5 is len('file:')
        return path

class SingleSourcesList(SourcesList):
    def __init__(self, file):
        self.matcher = SourceEntryMatcher("/usr/share/update-manager/channels/")
        self.list = []
        self.load(file)



#!/usr/bin/python3

"""Test cases for softwareproperties.AptAuth class."""

import os
import subprocess
import time
import unittest
import unittest.mock

import softwareproperties.AptAuth

KINETIC_DEFAULT = """\
tru:t:1:1658826897:0:3:1:5
pub:-:4096:1:D94AA3F0EFE21092:1336774248:::-:::scSC::::::23::0:
fpr:::::::::843938DF228D22F7B3742BC0D94AA3F0EFE21092:
uid:-::::1336774248::77355A0B96082B2694009775B6490C605BD16B6F::\
Ubuntu CD Image Automatic Signing Key (2012) <cdimage@ubuntu.com>::::::::::0:
pub:-:4096:1:871920D1991BC93C:1537196506:::-:::scSC::::::23::0:
fpr:::::::::F6ECB3762474EDA9D21B7022871920D1991BC93C:
uid:-::::1537196506::BE438F08F546424C0EA810FD722053597EB5127B::\
Ubuntu Archive Automatic Signing Key (2018) <ftpmaster@ubuntu.com>::::::::::0:
"""

KINETIC_SQ = """\
0. 843938DF228D22F7B3742BC0D94AA3F0EFE21092 Ubuntu CD Image Automatic Signing Key (2012) <cdimage@ubuntu.com>
1. F6ECB3762474EDA9D21B7022871920D1991BC93C Ubuntu Archive Automatic Signing Key (2018) <ftpmaster@ubuntu.com>\
"""

class TestAptAuth(unittest.TestCase):
    """Test softwareproperties.AptAuth class."""

    @staticmethod
    def setUp():
        os.environ["TZ"] = "UTC"
        time.tzset()

    @unittest.mock.patch("os.listdir")
    @unittest.mock.patch("subprocess.run")
    def test_list_sq(self, run_mock, listdir_mock):
        """Test softwareproperties.AptAuth.list."""
        apt_auth = softwareproperties.AptAuth.AptAuth()
        listdir_mock.return_value = ["demo-key.gpg", "second-key.gpg"]

        run_mock.return_value = subprocess.CompletedProcess(
            args=[],
            returncode=0,
            stdout=KINETIC_SQ,
        )

        self.assertEqual(
            apt_auth.list(),
            [
                "/etc/apt/trusted.gpg.d/demo-key.gpg\n"
                "  0. 843938DF228D22F7B3742BC0D94AA3F0EFE21092 Ubuntu CD Image Automatic Signing Key (2012) <cdimage@ubuntu.com>\n"
                "  1. F6ECB3762474EDA9D21B7022871920D1991BC93C Ubuntu Archive Automatic Signing Key (2018) <ftpmaster@ubuntu.com>",
                "/etc/apt/trusted.gpg.d/second-key.gpg\n"
                "  0. 843938DF228D22F7B3742BC0D94AA3F0EFE21092 Ubuntu CD Image Automatic Signing Key (2012) <cdimage@ubuntu.com>\n"
                "  1. F6ECB3762474EDA9D21B7022871920D1991BC93C Ubuntu Archive Automatic Signing Key (2018) <ftpmaster@ubuntu.com>"
            ],
        )


    @unittest.mock.patch("os.listdir")
    @unittest.mock.patch("subprocess.run")
    def test_list_gpg(self, run_mock, listdir_mock):
        """Test softwareproperties.AptAuth.list."""
        apt_auth = softwareproperties.AptAuth.AptAuth()
        listdir_mock.return_value = ["demo-key.gpg", "second-key.gpg"]

        gpgv_ret = subprocess.CompletedProcess(
            args=[],
            returncode=0,
            stdout=KINETIC_DEFAULT,
        )

        times_called = 0
        def raise_side_effect(self, *args, **more):
            nonlocal times_called
            times_called += 1
            if times_called % 2 == 0:
                return gpgv_ret
            else:
                raise OSError()

        run_mock.side_effect = raise_side_effect
        self.assertEqual(
            apt_auth.list(),
            [
                "/etc/apt/trusted.gpg.d/demo-key.gpg\n"
                "  843938DF228D22F7B3742BC0D94AA3F0EFE21092 Ubuntu CD Image Automatic Signing Key (2012) <cdimage@ubuntu.com>\n"
                "  F6ECB3762474EDA9D21B7022871920D1991BC93C Ubuntu Archive Automatic Signing Key (2018) <ftpmaster@ubuntu.com>",
                "/etc/apt/trusted.gpg.d/second-key.gpg\n"
                "  843938DF228D22F7B3742BC0D94AA3F0EFE21092 Ubuntu CD Image Automatic Signing Key (2012) <cdimage@ubuntu.com>\n"
                "  F6ECB3762474EDA9D21B7022871920D1991BC93C Ubuntu Archive Automatic Signing Key (2018) <ftpmaster@ubuntu.com>"
            ],
        )


if __name__ == "__main__":
    unittest.main()



# dialog_apt_key.py.in - edit the apt keys
#
#  Copyright (c) 2006-2007 Canonical
#
#  Author: Michael Vogt <mvo@debian.org>
#          Sebastian Heinlein <glatzor@ubuntu.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import aptsources.distro
from gettext import gettext as _

def get_popcon_description(distro):
    if isinstance(distro, aptsources.distro.UbuntuDistribution):
        return(_("<i>To improve the user experience of Ubuntu please "
                 "take part in the popularity contest. If you do so the "
                 "list of installed software and how often it was used will "
                 "be collected and sent anonymously to the Ubuntu project "
                 "on a weekly basis.\n\n"
                 "The results are used to improve the support for popular "
                 "applications and to rank applications in the search "
                 "results.</i>"))
    elif isinstance(distro, aptsources.distro.DebianDistribution):
        return(_("<i>To improve the user experience of Debian please take "
                 "part in the popularity contest. If you do so the list of "
                 "installed software and how often it was used will be "
                 "collected and sent anonymously to the Debian project.\n\n"
                 "The results are used to optimise the layout of the "
                 "installation CDs.</i>"))
    else:
        return(_("Submit the list of installed software and how often it is "
                 "is used to the distribution project."))




# dialog_add.py.in - dialog to add a new repository
#
#  Copyright (c) 2006 FSF Europe
#
#  Authors:
#       Sebastian Heinlein <glatzor@ubuntu.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import os
import gi
gi.require_version("Gtk", "4.0")
from gi.repository import GObject, Gtk, Gio, GLib
from gettext import gettext as _
import threading

import softwareproperties.gtk.dialogs as dialogs
from softwareproperties.MirrorTest import MirrorTest
from softwareproperties.gtk.utils import setup_ui
from softwareproperties.CountryInformation import CountryInformation

(COLUMN_PROTO, COLUMN_DIR) = range(2)

class MirrorItem(GObject.GObject):
    __gtype_name__ = 'MirrorItem'

    name = GObject.Property(type=str, default='')

    def __init__(self, name, mirror=None, is_country=False):
        super().__init__()
        self.props.name = name
        self.mirror = mirror
        self.is_country = is_country
        self.children = Gio.ListStore.new(MirrorItem) if is_country else None

class DialogMirror:
    def __init__(self, parent, datadir, distro):
        setup_ui(self, os.path.join(datadir, "gtkbuilder", "dialog-mirror.ui"),
                 domain="software-properties")

        self.dialog = self.dialog_mirror
        self.dialog.set_transient_for(parent)

        self.dialog_test = self.dialog_mirror_test
        self.dialog_test.set_transient_for(self.dialog)

        self.distro = distro
        self._setup_ui_references()
        self._setup_protocol_combo()
        self._setup_mirror_tree()
        self._connect_signals()

    def _setup_ui_references(self):
        self.listview = self.listview_mirrors
        self.button_choose = self.button_mirror_choose
        self.button_cancel = self.button_test_cancel
        self.label_test = self.label_test_mirror
        self.progressbar_test = self.progressbar_test_mirror
        self.combobox = self.combobox_mirror_proto

    def _setup_protocol_combo(self):
        model_proto = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        self.combobox.set_model(model_proto)

        renderer = Gtk.CellRendererText()
        self.combobox.pack_start(renderer, True)
        self.combobox.add_attribute(renderer, "markup", 0)

    def _setup_mirror_tree(self):
        root_store = self._build_mirror_tree()

        self.treelist = Gtk.TreeListModel.new(
            root_store, False, True, self._get_children_func)
        self.selection = Gtk.SingleSelection.new(self.treelist)
        self.listview.set_model(self.selection)

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup)
        factory.connect("bind", self._on_factory_bind)
        self.listview.set_factory(factory)

    def _build_mirror_tree(self):
        """Build the hierarchical mirror data structure. The mirror data are kept in python-apt-common
        at /usr/share/python-apt/templates/ubuntu.mirrors"""
        country_info = CountryInformation()
        countries = {}  # ISO code -> MirrorItem
        orphaned_mirrors = []

        # Group mirrors by country
        for hostname, mirror in self.distro.source_template.mirror_set.items():
            mirror_item = MirrorItem(hostname, mirror, is_country=False)

            country_code = getattr(mirror, 'location', None)
            if country_code:
                if country_code not in countries:
                    country_name = country_info.get_country_name(country_code)
                    countries[country_code] = MirrorItem(country_name, is_country=True)
                countries[country_code].children.append(mirror_item)
            else:
                orphaned_mirrors.append(mirror_item)

        for country_item in countries.values():
            if len(country_item.children) > 1:
                sorter = Gtk.StringSorter.new(Gtk.PropertyExpression.new(MirrorItem, None, "name"))
                sorted_children = Gtk.SortListModel.new(country_item.children, sorter)
                country_item.children = sorted_children

        root_store = Gio.ListStore.new(MirrorItem)

        for country_item in countries.values():
            root_store.append(country_item)

        for mirror_item in orphaned_mirrors:
            root_store.append(mirror_item)

        root_sorter = Gtk.StringSorter.new(Gtk.PropertyExpression.new(MirrorItem, None, "name"))
        sorted_root = Gtk.SortListModel.new(root_store, root_sorter)

        return sorted_root

    def _get_children_func(self, item):
        """TreeListModel child function"""
        return item.children

    def _on_factory_setup(self, factory, list_item):
        """Setup list item widget hierarchy"""
        expander = Gtk.TreeExpander()
        label = Gtk.Label(halign=Gtk.Align.START)
        expander.set_child(label)
        list_item.set_child(expander)

    def _on_factory_bind(self, factory, list_item):
        """Bind data to list item"""
        expander = list_item.get_child()
        label = expander.get_child()
        row = list_item.get_item()

        expander.set_list_row(row)
        item = row.get_item()
        label.set_text(item.props.name)

    def _connect_signals(self):
        """Connect all signal handlers"""
        self.button_mirror_test.connect("clicked", self._on_test_clicked)
        self.button_cancel.connect("clicked", self._on_cancel_test_clicked)
        self.selection.connect("notify::selected", self._on_selection_changed)

    def _on_selection_changed(self, selection, pspec):
        row = selection.get_selected_item()
        if not row:
            return

        item = row.get_item()
        self._update_protocol_combo(item)

    def _update_protocol_combo(self, item):
        model = self.combobox.get_model()
        model.clear()

        self.button_choose.set_sensitive(False)
        self.combobox.set_sensitive(False)

        # Only enable for actual mirrors (not countries)
        if item.is_country or not item.mirror:
            return

        # Populate protocols
        protocols = set()
        for repo in item.mirror.repositories:
            if repo.proto not in protocols:
                protocols.add(repo.proto)
                model.append(repo.get_info())

        if len(model) > 0:
            self.combobox.set_active(0)
            self.combobox.set_sensitive(True)
            self.button_choose.set_sensitive(True)

    def _get_selected_mirror_url(self):
        row = self.selection.get_selected_item()
        if not row:
            return None

        item = row.get_item()
        if item.is_country or not item.mirror:
            return None

        model = self.combobox.get_model()
        active = self.combobox.get_active()
        if active < 0:
            return None

        iter_proto = model.get_iter_from_string(str(active))
        proto = model.get_value(iter_proto, COLUMN_PROTO)
        directory = model.get_value(iter_proto, COLUMN_DIR)

        return f"{proto}://{item.mirror.hostname}/{directory}"

    def present(self, on_response):
        def _on_response(dialog, response_id):
            self.dialog.hide()
            url = None
            if response_id == Gtk.ResponseType.OK:
                url = self._get_selected_mirror_url()
            if callable(on_response):
                on_response(url)

        self.dialog.connect('response', _on_response)
        self.dialog.present()

    def _on_test_clicked(self, button):
        """Start mirror speed test"""
        self.button_cancel.set_sensitive(True)
        self.dialog_test.present()

        self.running = threading.Event()
        self.running.set()
        progress_update = threading.Event()

        # Get test parameters
        arch = os.popen("dpkg --print-architecture").read().strip()
        test_file = "dists/%s/%s/binary-%s/Packages.gz" % (
            self.distro.source_template.name,
            self.distro.source_template.components[0].name,
            arch,
        )

        # Start test
        mirrors = list(self.distro.source_template.mirror_set.values())
        test = MirrorTest(mirrors, test_file, progress_update, self.running)
        test.start()

        # Monitor progress
        GLib.timeout_add(100, self._update_test_progress, test, progress_update)

    def _update_test_progress(self, test, progress_update):
        if progress_update.is_set():
            progress_text = _("Completed %s of %s tests") % (test.progress[0], test.progress[1])
            self.progressbar_test.set_text(progress_text)
            self.progressbar_test.set_fraction(test.progress[2])
            progress_update.clear()

        if self.running.is_set():
            return True  # Continue updating

        # Test completed
        self.dialog_test.hide()
        self.label_test.set_label("")

        if test.best:
            self._select_best_mirror(test.best)
        else:
            dialogs.show_error_dialog(
                self.dialog,
                _("No suitable download server was found"),
                _("Please check your Internet connection.")
            )
        return False  # Stop updating

    def _select_best_mirror(self, best_hostname):
        for i in range(self.treelist.get_n_items()):
            row = self.treelist.get_item(i)
            item = row.get_item()

            if not item.is_country and item.mirror and item.mirror.hostname == best_hostname:
                self.selection.set_selected(i)
                self._update_protocol_combo(item)
                self.listview.scroll_to(i, Gtk.ListScrollFlags.FOCUS, None)
                return

    def _on_cancel_test_clicked(self, button):
        self.running.clear()
        self.label_test.set_label(f"<i>{_('Canceling...')}</i>")
        self.button_cancel.set_sensitive(False)
        self.progressbar_test.set_fraction(1.0)

    def on_dialog_mirror_test_delete_event(self, dialog, event, data=None):
        self._on_cancel_test_clicked(None)
        return True



#  Copyright (c) 2023 Canonical
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import os
import gi
gi.require_version("Gtk", "4.0")
from gi.repository import Gtk

from aptsources.sourceslist import Deb822SourceEntry
from softwareproperties.gtk.utils import setup_ui


class _CustomSourceEntryField(Gtk.Grid):
    def __init__(self, window, input_changed_callback, key=None, value=None):
        super().__init__()
        self.window = window

        self.entry_key = Gtk.Entry(name="field_key", placeholder_text="Key")
        self.entry_val = Gtk.Entry(name="field_value", placeholder_text="Value")
        self.button_delete = Gtk.Button(icon_name="edit-delete")

        if key is not None:
            self.entry_key.set_text(key)

        if value is not None:
            self.entry_val.set_text(value)

        self.set_column_spacing(12)
        self.attach(self.entry_key, 0, 0, 1, 1)
        self.attach(self.entry_val, 1, 0, 1, 1)
        self.attach(self.button_delete, 2, 0, 1, 1)

        self.connect("destroy", input_changed_callback)
        self.entry_key.connect("changed", input_changed_callback)
        self.entry_val.connect("changed", input_changed_callback)
        self.button_delete.connect("clicked", self._button_del_field_clicked_callback)

    def _button_del_field_clicked_callback(self, button):
        # Find the row the delete button belongs to
        row = button.get_ancestor(Gtk.ListBoxRow)
        if row is None:
            return

        container = row.get_parent()
        if container:
            container.remove(row)   # This will also unparent the widget
            row = None              # Drop reference so we never touch it again

        self.window.set_default_size(self.window.get_allocated_width(), -1)
        self.window.queue_resize()

    def get_key(self):
        return self.entry_key.get_text()

    def get_val(self):
        return self.entry_val.get_text()


class DialogEditDeb822:
    def __init__(self, parent, sourceslist, source_entry, datadir):
        self.sourceslist = sourceslist
        self.source_entry = source_entry
        self.new_source_entry = None

        setup_ui(
            self,
            os.path.join(datadir, 'gtkbuilder', 'dialog-edit-deb822-source.ui'),
            domain='software-properties'
        )
        self.main = self.dialog_edit_deb822
        if parent is not None:
            self.main.set_transient_for(parent)

        self._populate_with_existing_source_entry()
        self._connect_signal_callbacks()

        self.button_edit_ok.connect(
            "clicked", lambda *_: self.main.response(Gtk.ResponseType.OK)
        )
        self.button_edit_cancel.connect(
            "clicked", lambda *_: self.main.response(Gtk.ResponseType.CANCEL)
        )
        try:
            self.main.set_default_widget(self.button_edit_ok)
        except AttributeError:
            pass
        for entry in (self.entry_types, self.entry_uris, self.entry_suites,
                      self.entry_comps, self.entry_comment):
            entry.connect("activate", self._activate_ok_if_valid)

        # Initial validation state
        self._input_changed_callback()

    def _activate_ok_if_valid(self, *_args):
        if self._is_current_input_valid():
            self.main.response(Gtk.ResponseType.OK)

    def _populate_with_existing_source_entry(self):
        self.entry_types.set_text(" ".join(self.source_entry.types))
        self.entry_uris.set_text(" ".join(self.source_entry.uris))
        self.entry_suites.set_text(" ".join(self.source_entry.suites))
        self.entry_comps.set_text(" ".join(self.source_entry.comps))
        self.entry_comment.set_text(self.source_entry.comment.rstrip())

        for key, value in self.source_entry.section.tags.items():
            if key not in ["Types", "URIs", "Suites", "Components"]:
                self._append_custom_source_entry_field(key=key, value=value)

    def _append_custom_source_entry_field(self, key=None, value=None):
        row = Gtk.ListBoxRow(selectable=False, activatable=False)
        field = _CustomSourceEntryField(self.main, self._input_changed_callback, key, value)
        row.set_child(field)
        self.list_box_additional_fields.append(row)

    def _get_custom_source_entry_fields(self):
        rows = []
        row = self.list_box_additional_fields.get_first_child()
        while row:
            rows.append(row)
            row = row.get_next_sibling()

        fields = [r.get_child() for r in rows]

        return [f for f in fields if any((f.get_key(), f.get_val()))]

    def _is_current_input_valid(self):
        if '' in (self.entry_types.get_text(),
                  self.entry_uris.get_text(),
                  self.entry_suites.get_text(),
                  self.entry_comps.get_text()):
            return False

        for field in self._get_custom_source_entry_fields():
            if '' in (field.get_key(), field.get_val()):
                return False

        return True

    def _connect_signal_callbacks(self):
        for entry in (self.entry_types, self.entry_uris, self.entry_suites,
                      self.entry_comps, self.entry_comment):
            entry.connect('changed', self._input_changed_callback)

        self.button_add_field.connect('clicked',
                                      self._button_add_field_clicked_callback)

    def _input_changed_callback(self, *data):
        self.button_edit_ok.set_sensitive(self._is_current_input_valid())

    def _button_add_field_clicked_callback(self, *data):
        self._append_custom_source_entry_field()

    def _source_entry_from_input(self):
        source_entry = Deb822SourceEntry(None, self.source_entry.file)

        source_entry.types = self.entry_types.get_text().split()
        source_entry.uris = self.entry_uris.get_text().split()
        source_entry.suites = self.entry_suites.get_text().split()
        source_entry.comps = self.entry_comps.get_text().split()
        source_entry.comment = self.entry_comment.get_text()

        for field in self._get_custom_source_entry_fields():
            source_entry.section[field.get_key()] = field.get_val()

        return source_entry

    def present(self, on_response):
        def _on_response(_dlg, rid):
            if rid == Gtk.ResponseType.OK:
                self.new_source_entry = self._source_entry_from_input()
            self.main.hide()
            if callable(on_response):
                on_response(rid)
        self.main.connect('response', _on_response)
        self.main.present()



# dialog_cache_outdated.py - inform the user to update the apt cache
#
#  Copyright (c) 2006 Canonical
#
#  Authors:
#       Sebastian Heinlein <sebastian.heinlein@web.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import os
import gi
gi.require_version("Gdk", "4.0")
gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("PackageKitGlib", "1.0")
from gi.repository import Gtk, Adw
from gi.repository import PackageKitGlib as packagekit
from gettext import gettext as _

from softwareproperties.gtk.utils import (
    setup_ui,
)


class ProgressDialog(Gtk.Window):
    """A small helper window to display progress"""

    def __init__(self, parent):
        Gtk.Window.__init__(self, title=_("Cache Refresh"))
        self.set_transient_for(parent)
        self.set_modal(True)
        self.set_deletable(False)

        self.box = Gtk.Box(spacing=6, orientation=Gtk.Orientation.VERTICAL)
        self.box.set_margin_top(12)
        self.box.set_margin_bottom(12)
        self.box.set_margin_start(12)
        self.box.set_margin_end(12)
        self.set_child(self.box)

        self.label = Gtk.Label(xalign=0)
        self.label.set_markup("<b><big>{}</big></b>".format(_("Refreshing software cache")))
        self.box.append(self.label)

        # create a progress bar
        self.progressbar = Gtk.ProgressBar()
        self.box.append(self.progressbar)


class DialogCacheOutdated:
    def __init__(self, parent, on_done, datadir):
        """setup up the gtk dialog"""
        self.parent = parent
        self.on_done = on_done

        setup_ui(self, os.path.join(datadir, "gtkbuilder", "dialog-cache-outofdate.ui"), domain="software-properties")
        self.dialog = self.dialog_cache_outofdate
        self.dialog.set_transient_for(parent)

    def _report_exception_and_finish(self, message, detail):
        d = Adw.MessageDialog.new(self.parent, str(message), str(detail))
        d.add_response("ok", _("_OK"))
        d.set_close_response("ok")
        d.connect("response", lambda d, r: (self.on_done() if callable(self.on_done) else None))
        d.present()

    def on_pktask_progress(self, progress, ptype, udata=(None,)):
        if ptype == packagekit.ProgressType.PERCENTAGE:
            perc = progress.get_property('percentage')
            self._pdia.progressbar.set_fraction(perc / 100.0)

    def on_pktask_finish(self, source, result, udata=(None,)):
        try:
            self._pktask.generic_finish(result)
        except Exception as e:
            self._pdia.hide()
            self.parent.set_sensitive(True)
            self._report_exception_and_finish(_("Error while refreshing cache"), str(e))
            return

        self._pdia.hide()
        self.parent.set_sensitive(True)
        if callable(self.on_done):
            self.on_done()

    def present(self):
        def _on_response(dialog, response_id):
            dialog.hide()
            if response_id == Gtk.ResponseType.APPLY:
                self._pktask = packagekit.Task()
                self._pdia = ProgressDialog(self.parent)
                self._pdia.present()
                self.parent.set_sensitive(False)
                try:
                    self._pktask.refresh_cache_async(
                        False,              # force
                        None,               # cancellable
                        self.on_pktask_progress,
                        (None,),            # progress data
                        self.on_pktask_finish,
                        (None,),            # finish data
                    )
                except Exception as e:
                    self._pdia.hide()
                    self.parent.set_sensitive(True)
                    self._report_exception_and_finish(_("Error while refreshing cache"), str(e))
            else:
                # User chose to close without reload; proceed to quit
                if callable(self.on_done):
                    self.on_done()

        self.dialog.set_modal(True)
        self.dialog.connect("response", _on_response)
        self.dialog.present()



"""
 SimpleGladeApp.py
 Module that provides an object oriented abstraction to pygtk and gtkbuilder
 Copyright (C) 2009 Michael Vogt
 based on ideas from SimpleGladeBuilder by Sandino Flores Moreno
"""

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; version 3.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA

import logging
import gi
gi.require_version("Gtk", "4.0")
from gi.repository import Gtk, Gio
from softwareproperties.gtk.utils import setup_ui
LOG=logging.getLogger(__name__)

# based on SimpleGladeApp
class SimpleGtkbuilderApp(Gtk.Application):

    def __init__(self, path, domain):
        Gtk.Application.__init__(self, application_id="com.ubuntu.SoftwareProperties",
                                 flags=Gio.ApplicationFlags.FLAGS_NONE)
        setup_ui(self, path, domain)
        self.connect("activate", self.on_activate)

    def on_activate(self, data=None):
        self.add_window(self.window_main)

        if not self.window_main.is_visible():
            self.window_main.show()

    def run(self):
        """
        Starts the main loop of processing events checking for Control-C.

        The default implementation checks wheter a Control-C is pressed,
        then calls on_keyboard_interrupt().

        Use this method for starting programs.
        """
        try:
            Gtk.Application.run(self, None)
        except KeyboardInterrupt:
            self.on_keyboard_interrupt()

    def on_keyboard_interrupt(self):
        """
        This method is called by the default implementation of run()
        after a program is finished by pressing Control-C.
        """
        pass

    def quit(self):
        """
        Quit processing events.
        The default implementation calls Gtk.main_quit()

        Useful for applications that needs a non gtk main loop.
        For example, applications based on gstreamer needs to override
        this method with gst.main_quit()
        """
        Gtk.Application.quit(self)




#!/usr/bin/python3

import os
import re
import shutil
import subprocess
import tempfile
import unittest

from aptsources.distro import get_distro
from aptsources.sourceslist import (SourceEntry, Deb822SourceEntry)
from copy import copy
from pathlib import Path
from types import SimpleNamespace
from urllib.request import urlopen
from urllib.error import (URLError, HTTPError)

from softwareproperties.cloudarchive import RELEASE_MAP
from softwareproperties.sourceutils import replace_source_entry


CODENAME = get_distro().codename

ADD_APT_REPOSITORY = shutil.which('add-apt-repository', path=f'.:{os.environ.get("PATH")}')

FAKE_URI = 'http://fake.archive.test.com/ubuntu'
SOURCESLIST_TEXT = f'deb {FAKE_URI} {CODENAME} main'

# types
TYPE_BINARY = 'deb'
TYPE_SOURCE = 'deb-src'
# pockets
RELEASE = 'release'
PROPOSED = 'proposed'
UPDATES = 'updates'
SECURITY = 'security'
BACKPORTS = 'backports'
# components
MAIN = frozenset(['main'])
UNIVERSE = frozenset(['universe'])
RESTRICTED = frozenset(['restricted'])
MULTIVERSE = frozenset(['multiverse'])

ARCHIVE_URI = 'http://fake.mirror.private.com/ubuntu/'
ARCHIVE_FILENAME_URI = re.sub(r'[:/.]+', '_', ARCHIVE_URI)
ARCHIVE_FILENAME = f'archive_uri-{ARCHIVE_FILENAME_URI}-{CODENAME}.list'

PPA_NAME = 'ubuntu-support-team/software-properties-autopkgtest'
PPA_URI = f'https://ppa.launchpadcontent.net/{PPA_NAME}/ubuntu/'
PPA_FILENAME = f'ubuntu-support-team-ubuntu-software-properties-autopkgtest-{CODENAME}.sources'
PPA_TRUSTED_FINGERPRINT = 'A17A D76F CBB7 A7D5 73C4  DF43 1E03 2FCE 2F88 6048'

ADD_APT_REPOSITORY_LONG_PARAMS = SimpleNamespace(
    remove='--remove',
    enable_source='--enable-source',
    component='--component',
    pocket='--pocket',
    yes='--yes',
    no_update='--no-update',
    ppa='--ppa',
    cloud='--cloud',
    uri='--uri',
    sourceslist='--sourceslist',
)

ADD_APT_REPOSITORY_SHORT_PARAMS = SimpleNamespace(
    remove='-r',
    enable_source='-s',
    component='-c',
    pocket='-p',
    yes='-y',
    no_update='-n',
    ppa='-P',
    cloud='-C',
    uri='-U',
    sourceslist='-S',
)


def has_network():
    try:
        with urlopen("https://launchpad.net/"):
            pass
    except (URLError, HTTPError):
        return False
    return True

def line(uri=None, binary=True, pocket=None, components=None):
    deb = TYPE_BINARY if binary else TYPE_SOURCE
    uri = uri or FAKE_URI
    pocket = f'-{pocket}' if pocket else ''
    components = ' '.join(components or MAIN)
    return f'{deb} {uri} {CODENAME}{pocket} {components}'

def deb822_lines(uri=None, binary=True, source=False, pocket=None, components=None):
    uris = uri or FAKE_URI

    suites = f'{CODENAME}' + (f'-{pocket}' if pocket else '')

    types = [TYPE_BINARY] if binary else []
    types += [TYPE_SOURCE] if source else []
    types = ' '.join(types)

    components = components or MAIN
    components = ' '.join(components)

    return [f'Types: {types}',f'URIs: {uris}', f'Suites: {suites}', f'Components: {components}']


@unittest.skipIf(ADD_APT_REPOSITORY is None, 'Could not find add-apt-repository script')
class TestAddAptRepository(unittest.TestCase):
    params = ADD_APT_REPOSITORY_LONG_PARAMS

    @classmethod
    def setUpClass(cls):
        # test apt.conf file
        d = tempfile.TemporaryDirectory()
        cls.addClassCleanup(d.cleanup)
        apt_conf = Path(d.name) / 'apt.conf'

        # set APT_CONFIG to the test apt.conf
        test_env = copy(os.environ)
        test_env['APT_CONFIG'] = str(apt_conf)

        # dir for test apt-get script
        d = tempfile.TemporaryDirectory()
        cls.addClassCleanup(d.cleanup)
        apt_get_dir = Path(d.name)

        # create a test apt-get script and prepend it to the PATH
        apt_get_output = apt_get_dir / 'test-apt-get-output'
        apt_get_script = apt_get_dir / 'apt-get'
        apt_get_script.touch(mode=0o755)
        apt_get_script.write_text(f'''#!/bin/bash\necho "$*" >> {apt_get_output}\n''')
        test_env['PATH'] = f'{apt_get_dir}:{test_env["PATH"]}'

        cls.apt_conf = apt_conf
        cls.apt_get_output = apt_get_output
        cls.test_env = test_env

    def setUp(self):
        d = tempfile.TemporaryDirectory()
        self.addCleanup(d.cleanup)
        dir_etc = Path(d.name)
        self.apt_conf.write_text(f'Dir::Etc "{dir_etc}";')

        dir_etc.joinpath('apt.conf.d').mkdir(exist_ok=True)

        self.sourceslist = dir_etc / 'sources.list'
        self.sourceslist.touch()
        self.sourceslist.write_text(SOURCESLIST_TEXT)

        self.sourceslistd = dir_etc / 'sources.list.d'
        self.sourceslistd.mkdir(exist_ok=True)

        self.trustedgpgd = dir_etc / 'trusted.gpg.d'
        self.trustedgpgd.mkdir(exist_ok=True)

        self._sourcesfile = None
        self._trustedfile = None

        self.test_env.pop('LC_ALL', None)

        self.testing_ppa = False

        self.testing_uca = False
        self.test_env.pop('CA_ALLOW_CODENAME', None)

        self.noupdate = False
        self.pocket = None
        self.components = None
        self.enable_source = 0

    @property
    def sourcesfile(self):
        return self._sourcesfile

    @property
    def trustedfile(self):
        return self._trustedfile

    @sourcesfile.setter
    def sourcesfile(self, filename):
        self._sourcesfile = self._file_setter(self.sourceslistd, filename)

    @trustedfile.setter
    def trustedfile(self, filename):
        self._trustedfile = self._file_setter(self.trustedgpgd, filename)

    def _file_setter(self, d, f):
        if not f:
            return None
        path = d / f
        self.addCleanup(path.unlink, missing_ok=True)
        return path

    @property
    def parameter_enable_source(self):
        return [self.params.enable_source] * self.enable_source

    @property
    def parameter_pocket(self):
        return [self.params.pocket, self.pocket] if self.pocket else []

    @property
    def parameter_components(self):
        params = []
        for c in self.components or []:
            params.append(self.params.component)
            params.append(c)
        return params

    @property
    def parameter_all(self):
        return self.parameter_enable_source + self.parameter_pocket + self.parameter_components

    def line(self, uri=None, binary=True, pocket=None, components=None):
        return line(uri=uri or ARCHIVE_URI,
                    binary=binary,
                    pocket=pocket or self.pocket,
                    components=components or self.components)

    @unittest.skipUnless(os.geteuid() == 0, 'requires root')
    def call_add_apt_repository(self, args):
        self.apt_get_output.unlink(missing_ok=True)

        n = [self.params.no_update] if self.noupdate else []
        result = subprocess.run([ADD_APT_REPOSITORY] + ['-y'] + n + args,
                                stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                env=self.test_env)

        if result.returncode != 0:
            self.fail(f'add-apt-repository failed: {result.stdout}')

    def check_sourceentry(self, expected_line, content_lines):
        expected_entry = SourceEntry(expected_line)
        expected_entry.comps = set(expected_entry.comps)

        for (index, line) in enumerate(content_lines):
            entry = SourceEntry(line)
            entry.comps = set(entry.comps)
            if entry == expected_entry:
                return content_lines[:index] + content_lines[index+1:]

        content = '\n'.join(content_lines)
        self.fail(f"Missing line: '{expected_line}' in: '{content}'")

    def check_expected_lines(self, expected_lines, lines):
        for expected_line in expected_lines:
            lines = self.check_sourceentry(expected_line, lines)

        # ignore invalid or disabled lines
        lines = list(filter(lambda l: not SourceEntry(l).invalid, lines))
        lines = list(filter(lambda l: not SourceEntry(l).disabled, lines))
        if lines:
            unexpected_lines = '\n'.join(lines)
            self.fail(f"Unexpected lines:\n{unexpected_lines}")

    def check_expected_deb822_lines(self, expected_lines, lines):
        def filter_lines(lines):
            in_ignored_field = False
            for line in lines:
                if line.startswith('Signed-By:'):
                    in_ignored_field = True
                    continue

                if in_ignored_field and line.startswith(' '):
                    continue

                in_ignored_field = False
                yield line

        # Ignore lines that should be there, but we are not checking for, e.g.
        # Signed-By.
        lines = filter_lines(lines)
        missing_lines = list(set(expected_lines) - set(lines))
        unexpected_lines = list(set(lines) - set(expected_lines))


        if missing_lines:
            self.fail(f'Missing lines: {missing_lines} in: {lines}')

        if unexpected_lines:
            self.fail(f'Unexpected lines: {unexpected_lines} in: {lines}')

    def apt_get_output_lines(self):
        try:
            return self.apt_get_output.read_text().splitlines()
        except FileNotFoundError:
            return []

    def check_apt_get_update(self):
        if self.noupdate:
            self.assertNotIn('update', self.apt_get_output_lines())
        else:
            self.assertIn('update', self.apt_get_output_lines())

    def check_uca_keyring(self):
        if self.testing_uca:
            self.assertIn('install -y ubuntu-cloud-keyring', self.apt_get_output_lines())

    def check_ppa_gpgkey(self):
        if not self.testing_ppa:
            return

        with open(self.sourcesfile, mode='r', encoding='utf-8') as f:
            entry = Deb822SourceEntry(f.read(), '')

        keydata = entry.section.get('Signed-By')
        keydata = keydata.splitlines()
        keydata = [l.lstrip() for l in keydata]
        keydata = [l if l != '.' else '' for l in keydata]
        keydata = '\n'.join(keydata)

        with tempfile.NamedTemporaryFile(mode='w') as f:
            f.write(keydata)
            f.flush()

            with tempfile.TemporaryDirectory() as homedir:
                Path(homedir).chmod(0o700) # so gpg doesn't complain to stderr
                cmd = f'gpg -q --homedir {homedir} --import {f.name}'
                subprocess.check_call(cmd.split())
                keyring = os.path.join(homedir, 'pubring.kbx')
                cmd = f'gpg -q --homedir {homedir} --no-default-keyring --keyring {keyring} --fingerprint'
                result = subprocess.run(cmd.split(), stdout=subprocess.PIPE, encoding='UTF-8')
                self.assertIn(PPA_TRUSTED_FINGERPRINT, result.stdout)

    def check_file(self, testfile, exists=True):
        if not testfile:
            return
        msg = f'{"missing" if exists else "unexpected"} {testfile}'
        self.assertEqual(testfile.exists(), exists, msg)

    def run_sourceslistd_add_remove_test(self, args, expected_lines=[]):
        self.check_file(self.sourcesfile, exists=False)
        self.check_file(self.trustedfile, exists=False)

        self.call_add_apt_repository(args)

        self.check_file(self.sourcesfile)
        self.check_file(self.trustedfile)

        if self.testing_ppa:
            self.check_expected_deb822_lines(expected_lines, self.sourcesfile.read_text().splitlines())
        else:
            self.check_expected_lines(expected_lines, self.sourcesfile.read_text().splitlines())

        self.check_apt_get_update()
        self.check_ppa_gpgkey()
        self.check_uca_keyring()

        self.call_add_apt_repository([self.params.remove] + args)

        self.check_file(self.sourcesfile, exists=False)

        # trusted.gpg.d file isn't removed by add-apt-repository
        self.check_file(self.trustedfile)

    def run_sourceslistd_test(self, args, expected_line):
        if self.testing_ppa:
            expected_lines = deb822_lines(uri=PPA_URI)
        else:
            source_entry = replace_source_entry(SourceEntry(expected_line), type=TYPE_SOURCE)
            source_entry.set_enabled(self.enable_source > 0)

            expected_lines = [expected_line, str(source_entry)]
        test_args = self.parameter_all
        test_args += args
        self.run_sourceslistd_add_remove_test(test_args, expected_lines)

    def test_uri(self):
        self.sourcesfile = ARCHIVE_FILENAME
        args = [self.params.uri, ARCHIVE_URI]
        self.run_sourceslistd_test(args, self.line())

    def test_uri_noupdate(self):
        self.noupdate = True
        self.test_uri()

    def test_sourceslist(self):
        self.sourcesfile = ARCHIVE_FILENAME
        line = self.line()
        args = [self.params.sourceslist, line]
        self.run_sourceslistd_test(args, line)

    def test_sourceslist_noupdate(self):
        self.noupdate = True
        self.test_sourceslist()

    @unittest.skipUnless(has_network(), 'requires network')
    def test_ppa(self, /, with_param=True, prefix=False):
        self.testing_ppa = True

        self.sourcesfile = PPA_FILENAME
        self.trustedfile = None
        args = []
        if with_param:
            args += [self.params.ppa]
        if prefix:
            args += [f'ppa:{PPA_NAME}']
        else:
            args += [PPA_NAME]
        self.run_sourceslistd_test(args, None)

    def test_ppa_noupdate(self):
        self.noupdate = True
        self.test_ppa()

    def test_ppa_with_prefix(self):
        self.test_ppa(prefix=True)

    def test_ppa_without_param_with_prefix(self):
        self.test_ppa(with_param=False, prefix=True)

    def _test_uca(self, uca_release, /, with_param, prefix):
        self.testing_uca = True
        if CODENAME != RELEASE_MAP.get(uca_release):
            self.test_env['CA_ALLOW_CODENAME'] = CODENAME

        self.sourcesfile = f'cloudarchive-{uca_release}.list'
        line = f'deb http://ubuntu-cloud.archive.canonical.com/ubuntu {CODENAME}-updates/{uca_release} main'
        args = []
        if with_param:
            args += [self.params.cloud]
        if prefix:
            args += [f'{prefix}:{uca_release}']
        else:
            args += [uca_release]
        self.run_sourceslistd_test(args, line)

    def test_uca(self, /, with_param=True, prefix=None):
        uca_releases = list(RELEASE_MAP.keys())
        self.assertGreater(len(uca_releases), 0)
        for uca_release in uca_releases:
            with self.subTest(uca_release=uca_release):
                self.setUp()
                self._test_uca(uca_release, with_param=with_param, prefix=prefix)
                self.tearDown()

    def test_uca_noupdate(self):
        self.noupdate = True
        self.test_uca()

    def test_uca_with_prefix_uca(self):
        self.test_uca(prefix='uca')

    def test_uca_with_prefix_uca_without_param(self):
        self.test_uca(with_param=False, prefix='uca')

    def test_uca_with_prefix_cloud_archive(self):
        self.test_uca(prefix='cloud-archive')

    def test_uca_with_prefix_cloud_archive_without_param(self):
        self.test_uca(with_param=False, prefix='cloud-archive')

    class TestLinesGroup(object):
        def __init__(self, /, test_pocket=None, test_components=frozenset(), repo_n=1):
            self.test_pocket = test_pocket
            self.test_components = test_components
            self.repo_n = repo_n

            self.lines = []
            self.expected_lines = []
            self.after_remove = []

        def next_repo(self):
            self.repo_n += 1

        def entry(self, /, binary=True, pocket=None, components=MAIN):
            return SourceEntry(self.line(binary=binary, pocket=pocket, components=components))

        def line(self, /, binary=True, pocket=None, components=MAIN):
            n = self.repo_n
            return line(uri=f'{FAKE_URI}/source{n}', binary=binary, pocket=pocket, components=components)

    def check_lines_group(self, args, group):
        self.sourceslist.write_text('\n'.join(group.lines))

        self.call_add_apt_repository(args)
        self.check_expected_lines(group.expected_lines,
                                  self.sourceslist.read_text().splitlines())


        self.call_add_apt_repository([self.params.remove] + args)
        self.check_expected_lines(group.after_remove,
                                  self.sourceslist.read_text().splitlines())

    def add_lines_comp(self, group, components, first_line=True, other_lines=frozenset()):
        group.lines.append(group.line(components=components))

        expected = copy(components)
        if first_line:
            # all new components will be added to (only) the first line
            expected |= group.test_components
            expected -= other_lines
        group.expected_lines.append(group.line(components=expected))

        after_remove = components - group.test_components
        if after_remove:
            group.after_remove.append(group.line(components=after_remove))

    def test_component(self):
        # with no self.components, we test the no-prefix case using MAIN
        test_components = copy(self.components or MAIN)
        args = self.parameter_components or list(test_components)

        group = self.TestLinesGroup(test_components=test_components)

        # source1, with two lines, one component on each
        self.add_lines_comp(group, MAIN, other_lines=UNIVERSE)
        self.add_lines_comp(group, UNIVERSE, first_line=False)

        # source2, one line with two components
        group.next_repo()
        self.add_lines_comp(group, MAIN | UNIVERSE)

        # source3, same as source2 but different components
        group.next_repo()
        self.add_lines_comp(group, RESTRICTED | MULTIVERSE)

        # source4, one line, one component
        group.next_repo()
        self.add_lines_comp(group, RESTRICTED)

        self.check_lines_group(args, group)

    def test_component_main(self):
        self.components = MAIN
        self.test_component()

    def test_component_main_noupdate(self):
        self.noupdate = True
        self.test_component_main()

    def test_component_universe(self):
        self.components = UNIVERSE
        self.test_component()

    def test_component_universe_noupdate(self):
        self.noupdate = True
        self.test_component_universe()

    def test_component_restricted(self):
        self.components = RESTRICTED
        self.test_component()

    def test_component_multiverse(self):
        self.components = MULTIVERSE
        self.test_component()

    def test_component_main_universe(self):
        self.components = MAIN | UNIVERSE
        self.test_component()

    def test_component_universe_multiverse(self):
        self.components = UNIVERSE | MULTIVERSE
        self.test_component()

    def test_component_all(self):
        self.components = MAIN | UNIVERSE | RESTRICTED | MULTIVERSE
        self.test_component()

    def test_component_all_except_main(self):
        self.components = UNIVERSE | RESTRICTED | MULTIVERSE
        self.test_component()

    def add_lines_pocket(self, group, pocket, components=MAIN, other_pockets=[], other_comps=frozenset(), first_line=True):
        line = group.line(pocket=pocket, components=components)
        group.lines.append(line)
        group.expected_lines.append(line)
        if (pocket or RELEASE) != group.test_pocket:
            group.after_remove.append(line)

    def _test_pocket(self):
        self.assertIsNotNone(self.pocket)

        args = self.parameter_pocket

        group = self.TestLinesGroup(test_pocket=self.pocket)

        # source1, with two lines, one pocket, one component on each
        self.add_lines_pocket(group, None, MAIN)
        self.add_lines_pocket(group, None, UNIVERSE)
        if self.pocket not in [RELEASE]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | UNIVERSE))

        # source2, with two lines, two pockets, varying components on each
        group.next_repo()
        self.add_lines_pocket(group, None, MAIN | UNIVERSE)
        self.add_lines_pocket(group, UPDATES, MAIN)
        if self.pocket not in [RELEASE, UPDATES]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | UNIVERSE))

        # source3, with three lines, two pockets, varying components on each
        group.next_repo()
        self.add_lines_pocket(group, None, MAIN)
        self.add_lines_pocket(group, None, UNIVERSE)
        self.add_lines_pocket(group, UPDATES, MAIN)
        if self.pocket not in [RELEASE, UPDATES]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | UNIVERSE))

        # source4, no release pocket
        group.next_repo()
        self.add_lines_pocket(group, UPDATES, MAIN)
        if self.pocket not in [UPDATES]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | RESTRICTED))

        # source5, all pockets exist
        group.next_repo()
        self.add_lines_pocket(group, None, MAIN)
        self.add_lines_pocket(group, UPDATES, MAIN)
        self.add_lines_pocket(group, PROPOSED, MAIN)
        self.add_lines_pocket(group, SECURITY, MAIN)

        # source6, release lines specified with and without pocket
        group.next_repo()
        self.add_lines_pocket(group, None, MAIN)
        self.add_lines_pocket(group, RELEASE, MULTIVERSE)
        self.add_lines_pocket(group, UPDATES, MAIN)
        if self.pocket not in [RELEASE, UPDATES]:
            group.expected_lines.append(group.line(pocket=self.pocket, components=MAIN | MULTIVERSE))

        self.check_lines_group(args, group)

    def test_pocket_release(self):
        self.pocket = RELEASE
        self._test_pocket()

    def test_pocket_release_noupdate(self):
        self.noupdate = True
        self.test_pocket_release()

    def test_pocket_updates(self):
        self.pocket = UPDATES
        self._test_pocket()

    def test_pocket_updates_noupdate(self):
        self.noupdate = True
        self.test_pocket_updates()

    def test_pocket_proposed(self):
        self.pocket = PROPOSED
        self._test_pocket()

    def test_pocket_security(self):
        self.pocket = SECURITY
        self._test_pocket()

    def add_lines_es(self, group, pocket, components, /, binary=True, disabled=False, alone=False, plus_comps=[], minus_comps=False):
        binary_entry = replace_source_entry(group.entry(pocket=pocket, components=components), disabled=disabled)
        enabled_source_entry = group.entry(binary=False, pocket=pocket, components=components)
        disabled_source_entry = replace_source_entry(enabled_source_entry, disabled=True)

        if binary:
            binary_line = str(binary_entry)
            group.lines.append(binary_line)
            group.expected_lines.append(binary_line)
            group.after_remove.append(binary_line)
            # with binary, 'alone' indicates no matching source line exists,
            # so we need to create one if enable_source > 1
            # if minus_comps is True, our comps will be added to another line
            if alone and not disabled and not minus_comps and self.enable_source > 1:
                group.expected_lines.append(str(enabled_source_entry))
                group.after_remove.append(str(disabled_source_entry))
        else:
            enabled_line = str(enabled_source_entry)
            disabled_line = str(disabled_source_entry)
            if not disabled:
                group.lines.append(enabled_line)
            else:
                group.lines.append(disabled_line)
            if plus_comps and self.enable_source > 1:
                # include comps from other binary-only lines
                comps = set(enabled_source_entry.comps) | set(plus_comps)
                enabled_line = str(replace_source_entry(enabled_source_entry, comps=comps))
                disabled_line = str(replace_source_entry(disabled_source_entry, comps=comps))
            if disabled and alone:
                # with source, 'alone' indicates no matching (enabled) binary line exists,
                # so the source line will *not* be enabled
                group.expected_lines.append(disabled_line)
            else:
                # if already enabled, or if not 'alone', we expect it to be enabled
                group.expected_lines.append(enabled_line)
            group.after_remove.append(disabled_line)

    def add_lines_es_binary(self, group, pocket, components, /, **kwargs):
        self.add_lines_es(group, pocket, components, binary=True, **kwargs)

    def add_lines_es_source(self, group, pocket, components, /, **kwargs):
        self.add_lines_es(group, pocket, components, binary=False, **kwargs)

    def add_lines_es_both(self, group, pocket, components, /, source_disabled=False, **kwargs):
        self.add_lines_es_binary(group, pocket, components, **kwargs)
        if kwargs.get('disabled'):
            # if both lines are disabled, source line is effectively 'alone'
            # as there is no enabled matching binary line
            kwargs['alone'] = True
        if source_disabled:
            kwargs['disabled'] = True
        self.add_lines_es_source(group, pocket, components, **kwargs)

    def _test_enable_source(self):
        self.assertGreater(self.enable_source, 0)

        args = self.parameter_enable_source

        group = self.TestLinesGroup()

        # source1, simple test with just binary lines
        self.add_lines_es_binary(group, RELEASE, MAIN | UNIVERSE, alone=True)
        self.add_lines_es_binary(group, UPDATES, MAIN | UNIVERSE, alone=True)
        self.add_lines_es_binary(group, PROPOSED, MAIN | UNIVERSE, alone=True)
        self.add_lines_es_binary(group, SECURITY, MAIN | UNIVERSE, alone=True)

        # source2, all with matching disabled source lines
        group.next_repo()
        self.add_lines_es_both(group, None, MAIN | UNIVERSE, source_disabled=True)
        self.add_lines_es_both(group, None, RESTRICTED | MULTIVERSE, source_disabled=True)
        self.add_lines_es_both(group, UPDATES, MAIN | UNIVERSE | RESTRICTED | MULTIVERSE, source_disabled=True)
        self.add_lines_es_both(group, SECURITY, MAIN | UNIVERSE, source_disabled=True)

        # source3, mix of just binary lines plus some disabled and enabled source lines
        group.next_repo()
        self.add_lines_es_both(group, None, MAIN | UNIVERSE, source_disabled=True, plus_comps=RESTRICTED | MULTIVERSE)
        self.add_lines_es_binary(group, None, RESTRICTED | MULTIVERSE, alone=True, minus_comps=True)
        self.add_lines_es_both(group, UPDATES, MAIN | UNIVERSE | RESTRICTED | MULTIVERSE, source_disabled=True)
        self.add_lines_es_both(group, SECURITY, MAIN | UNIVERSE)

        # source4, disabled binary lines to verify disabled source lines aren't enabled
        group.next_repo()
        self.add_lines_es_both(group, None, MAIN | UNIVERSE, disabled=True)
        self.add_lines_es_both(group, UPDATES, MULTIVERSE, disabled=True)

        self.check_lines_group(args, group)

    def test_enable_source_1(self):
        self.enable_source = 1
        self._test_enable_source()

    def test_enable_source_1_noupdate(self):
        self.no_update = True
        self.test_enable_source_1()

    def test_enable_source_2(self):
        self.enable_source = 2
        self._test_enable_source()

    def test_enable_source_2_noupdate(self):
        self.no_update = True
        self.test_enable_source_2()

    @classmethod
    def add_locale_test(cls, testname, locale):
        newname = f"{testname}_locale_{re.sub(r'[-.]+', '_', locale)}"
        if hasattr(cls, newname):
            return
        def _test_locale(self):
            self.test_env['LC_ALL'] = locale
            getattr(self, testname)()
        setattr(cls, newname, _test_locale)

    @classmethod
    def add_combination_test(cls, testname, /, enable_source=0, pocket=None, components=None):
        newname = testname
        if enable_source > 0:
            newname += f'_enable_source_{enable_source}'
        if pocket:
            newname += f'_pocket_{pocket}'
        if components:
            newname += f'_components_{"_".join(components)}'
        if hasattr(cls, newname):
            return
        def _test_combination(self):
            self.enable_source = enable_source
            self.pocket = pocket
            self.components = components
            getattr(self, testname)()
        setattr(cls, newname, _test_combination)


class TestAddAptRepositoryShortParams(TestAddAptRepository):
    params = ADD_APT_REPOSITORY_SHORT_PARAMS


# Add various combination tests
for test in ['uri', 'sourceslist']:
    for locale in ['C', 'C.UTF-8']:
        TestAddAptRepository.add_locale_test(f'test_{test}', locale=locale)
    for enable_source in [0, 1, 2]:
        for pocket in [None, RELEASE, UPDATES, PROPOSED, SECURITY]:
            for components in [None, MAIN, UNIVERSE, RESTRICTED, MULTIVERSE, MAIN | UNIVERSE,
                               MAIN | UNIVERSE | RESTRICTED | MULTIVERSE]:
                TestAddAptRepository.add_combination_test(f'test_{test}',
                                                          enable_source=enable_source,
                                                          pocket=pocket,
                                                          components=components)


if __name__ == "__main__":
    unittest.main(verbosity=2)



#!/usr/bin/python3

import apt

from functools import partial
import unittest
import sys
import os

from aptsources.distro import get_distro
from aptsources.sourceslist import (SourceEntry, Deb822SourceEntry)
from contextlib import contextmanager
from http.client import HTTPException
from launchpadlib.launchpad import Launchpad
from lazr.restfulclient.errors import Unauthorized
from urllib.request import urlopen
from urllib.error import URLError
from unittest.mock import (patch, Mock)

sys.path.insert(0, "..")

from softwareproperties.sourceslist import SourcesListShortcutHandler
from softwareproperties.uri import URIShortcutHandler
from softwareproperties.cloudarchive import CloudArchiveShortcutHandler
from softwareproperties.ppa import PPAShortcutHandler
from softwareproperties.shortcuthandler import InvalidShortcutException, ShortcutException
from softwareproperties.shortcuts import shortcut_handler


DISTRO = get_distro()
CODENAME = DISTRO.codename

# These must match the ppa used in the VALID_PPAS
PPA_LINE = f"""Types: deb
URIs: https://ppa.launchpadcontent.net/ddstreet/ppa/ubuntu/
Suites: {CODENAME}
Components: main
"""
PPA_FILEBASE = "ddstreet-ubuntu-ppa"
PPA_SOURCEFILE = f"{PPA_FILEBASE}-{CODENAME}.sources"
PPA_NETRCFILE = f"{PPA_FILEBASE}.conf"

PRIVATE_PPA_PASSWORD = "thisisnotarealpassword"
PRIVATE_PPA_LINE = f"""Types: deb
URIs: https://private-ppa.launchpadcontent.net/ddstreet/ppa/ubuntu/
Suites: {CODENAME}
Components: main
"""
PRIVATE_PPA_NETRCCONTENT = f"machine private-ppa.launchpadcontent.net/ddstreet/ppa/ubuntu/ login ddstreet password {PRIVATE_PPA_PASSWORD}"
PRIVATE_PPA_SUBSCRIPTION_URL = f"https://ddstreet:{PRIVATE_PPA_PASSWORD}@private-ppa.launchpadcontent.net/ddstreet/ppa/ubuntu/"

# These must match the uca used in VALID_UCAS
UCA_CANAME = "train"
UCA_ARCHIVE = "http://ubuntu-cloud.archive.canonical.com/ubuntu"
UCA_LINE = f"deb {UCA_ARCHIVE} {CODENAME}-updates/{UCA_CANAME} main"
UCA_LINE_PROPOSED = f"deb {UCA_ARCHIVE} {CODENAME}-proposed/{UCA_CANAME} main"
UCA_FILEBASE = f"cloudarchive-{UCA_CANAME}"
UCA_SOURCEFILE = f"{UCA_FILEBASE}.list"

# This must match the VALID_URIS
URI = "http://fake.mirror.private.com/ubuntu"
URI_FILEBASE = "archive_uri-http_fake_mirror_private_com_ubuntu"
URI_SOURCEFILE = f"{URI_FILEBASE}-{CODENAME}.list"

VALID_LINES = [f"deb {URI} {CODENAME} main"]
VALID_URIS = [URI]
VALID_PPAS = ["ppa:ddstreet", "ppa:~ddstreet", "ppa:ddstreet/ppa", "ppa:~ddstreet/ppa", "ppa:ddstreet/ubuntu/ppa", "ppa:~ddstreet/ubuntu/ppa"]
VALID_UCAS = [f"cloud-archive:{UCA_CANAME}", f"cloud-archive:{UCA_CANAME}-updates", f"cloud-archive:{UCA_CANAME}-proposed", f"uca:{UCA_CANAME}", f"uca:{UCA_CANAME}-updates", f"uca:{UCA_CANAME}-proposed"]
VALID_ALL = VALID_LINES + VALID_URIS + VALID_PPAS + VALID_UCAS

INVALID_LINES = ["xxx invalid deb line"]
INVALID_URIS = ["invalid"]
INVALID_PPAS = ["ppainvalid:ddstreet", "ppa:ddstreet/ubuntu/ppa/invalid"]
INVALID_UCAS = [f"cloud-invalid:{UCA_CANAME}", "cloud-archive:"]
INVALID_ALL = INVALID_LINES + INVALID_URIS + INVALID_PPAS + INVALID_UCAS

def has_network():
    try:
        with urlopen("https://launchpad.net/"):
            pass
    except (URLError, HTTPException):
        return False
    return True

def mock_login_with(*args, **kwargs):
    has_subscription = kwargs.pop("has_subscription", True)

    _lp = Launchpad.login_anonymously(*args, **kwargs)
    lp = Mock(wraps=_lp)

    lp.me = Mock()
    lp.me.name = 'ddstreet'

    def mock_getPPAByName(_team, name):
        _ppa = _team.getPPAByName(name=name)
        ppa = Mock(wraps=_ppa)
        ppa.signing_key_fingerprint = _ppa.signing_key_fingerprint
        ppa.private = True
        return ppa

    def mock_people(teamname):
        _team = _lp.people(teamname)
        team = Mock(wraps=_team)
        team.getPPAByName = lambda name: mock_getPPAByName(_team, name)
        team.getArchiveSubscriptionURL = lambda archive: PRIVATE_PPA_SUBSCRIPTION_URL
        if has_subscription:
            team.getArchiveSubscriptionURL = lambda archive: PRIVATE_PPA_SUBSCRIPTION_URL
        else:
            def raise_unauthorized(archive):
                raise Unauthorized(Mock(), Mock())

            team.getArchiveSubscriptionURL = raise_unauthorized

        return team

    lp.people = mock_people
    return lp


class ShortcutsTestcase(unittest.TestCase):
    enable_source = False

    @classmethod
    def setUpClass(cls):
        for k in apt.apt_pkg.config.keys():
            apt.apt_pkg.config.clear(k)
        apt.apt_pkg.init()

    def setUp(self):
        # avoid polution from other tests
        apt.apt_pkg.config.set("Dir", "/")
        apt.apt_pkg.config.set("Dir::Etc", "etc/apt")
        apt.apt_pkg.config.set("Dir::Etc::sourcelist", "sources.list")
        apt.apt_pkg.config.set("Dir::Etc::sourceparts", "sources.list.d")
        apt.apt_pkg.config.set("Dir::Etc::trustedparts", "trusted.gpg.d")
        apt.apt_pkg.config.set("Dir::Etc::netrcparts", "auth.conf.d")

    def create_handler(self, line, handler, *args, **kwargs):
        return handler(line, *args, enable_source=self.enable_source, **kwargs)

    def create_handlers(self, line, handler, *args, **kwargs):
        handlers = handler if isinstance(handler, list) else [handler]
        # note, always appends shortcut_handler
        return [self.create_handler(line, handler, *args, **kwargs)
                for handler in handlers + [shortcut_handler]]

    def check_shortcut(self, shortcut, line, sourcefile=None, trustedfile=None, netrcfile=None,
                       sourceparts=apt.apt_pkg.config.find_dir("Dir::Etc::sourceparts"),
                       trustedparts=apt.apt_pkg.config.find_dir("Dir::Etc::trustedparts"),
                       netrcparts=apt.apt_pkg.config.find_dir("Dir::Etc::netrcparts"),
                       trustedcontent=False, netrccontent=None):
        self.assertEqual(shortcut.SourceEntry().line, line)

        self.assertEqual(shortcut.sourceparts_path, sourceparts)
        if sourcefile:
            self.assertEqual(shortcut.SourceEntry().file, os.path.join(sourceparts, sourcefile))
            self.assertEqual(shortcut.sourceparts_filename, sourcefile)
            self.assertEqual(shortcut.sourceparts_file, os.path.join(sourceparts, sourcefile))

        if shortcut.deb822:
            binentry = Deb822SourceEntry(line, '')
            binentry.types = [DISTRO.binary_type]
        else:
            binentry = SourceEntry(line)
            binentry.type = DISTRO.binary_type

        self.assertEqual(shortcut.SourceEntry(shortcut.binary_type), binentry)

        if shortcut.deb822:
            srcentry = Deb822SourceEntry(line, '')
            srcentry.types = [DISTRO.source_type]
        else:
            srcentry = SourceEntry(line)
            srcentry.type = DISTRO.source_type

        srcentry.set_enabled(self.enable_source)
        self.assertEqual(shortcut.SourceEntry(shortcut.source_type), srcentry)

        self.assertEqual(shortcut.trustedparts_path, trustedparts)
        if trustedfile:
            self.assertEqual(shortcut.trustedparts_filename, trustedfile)
            self.assertEqual(shortcut.trustedparts_file, os.path.join(trustedparts, trustedfile))

        # Checking the actual gpg key content is too much work.
        if trustedcontent:
            self.assertIsNotNone(shortcut.trustedparts_content)
        else:
            self.assertIsNone(shortcut.trustedparts_content)

        self.assertEqual(shortcut.netrcparts_path, netrcparts)
        if netrcfile:
            self.assertEqual(shortcut.netrcparts_filename, netrcfile)
            self.assertEqual(shortcut.netrcparts_file, os.path.join(netrcparts, netrcfile))

        self.assertEqual(shortcut.netrcparts_content, netrccontent)

    def test_shortcut_sourceslist(self):
        for line in VALID_LINES:
            for shortcut in self.create_handlers(line, SourcesListShortcutHandler):
                self.check_shortcut(shortcut, line)

    def test_shortcut_uri(self):
        for uri in VALID_URIS:
            line = f"deb {uri} {CODENAME} main"
            for shortcut in self.create_handlers(uri, URIShortcutHandler):
                self.check_shortcut(shortcut, line, sourcefile=URI_SOURCEFILE)

    @unittest.skipUnless(has_network(), "requires network")
    def test_shortcut_ppa(self):
        for ppa in VALID_PPAS:
            for shortcut in self.create_handlers(ppa, PPAShortcutHandler):
                self.check_shortcut(shortcut, PPA_LINE,
                                    sourcefile=PPA_SOURCEFILE,
                                    netrcfile=PPA_NETRCFILE,
                                    trustedcontent=True)

    @unittest.skipUnless(has_network(), "requires network")
    def test_shortcut_private_ppa(self):
        # this is the same tests as the public ppa, but login=True will use the mocked lp instance
        # this *does not* actually test/verify this works with a real private ppa; that must be done manually
        with patch('launchpadlib.launchpad.Launchpad.login_with', new=mock_login_with):
            for ppa in VALID_PPAS:
                for shortcut in self.create_handlers(ppa, PPAShortcutHandler, login=True):
                    self.check_shortcut(shortcut, PRIVATE_PPA_LINE,
                                        sourcefile=PPA_SOURCEFILE,
                                        netrcfile=PPA_NETRCFILE,
                                        trustedcontent=True,
                                        netrccontent=PRIVATE_PPA_NETRCCONTENT)

    @unittest.skipUnless(has_network(), "requires network")
    def test_shortcut_private_ppa_no_subscription(self):
        # this is the same tests as the public ppa, but login=True will use the mocked lp instance
        # this *does not* actually test/verify this works with a real private ppa; that must be done manually
        with patch('launchpadlib.launchpad.Launchpad.login_with', new=partial(mock_login_with, has_subscription=False)):
            for ppa in VALID_PPAS:
                with self.assertRaises(ShortcutException):
                    self.create_handler(ppa, PPAShortcutHandler, login=True)

    @contextmanager
    def ca_allow_codename(self, codename):
        key = "CA_ALLOW_CODENAME"
        orig = os.environ.get(key, None)
        try:
            os.environ[key] = codename
            yield
        finally:
            if orig:
                os.environ[key] = orig
            else:
                os.environ.pop(key, None)

    def test_shortcut_cloudarchive(self):
        for uca in VALID_UCAS:
            line = UCA_LINE_PROPOSED if 'proposed' in uca else UCA_LINE
            with self.ca_allow_codename(CODENAME):
                for shortcut in self.create_handlers(uca, CloudArchiveShortcutHandler):
                    self.check_shortcut(shortcut, line, sourcefile=UCA_SOURCEFILE)

    def check_invalid_shortcut(self, handler, shortcut):
        msg = "'%s' should have rejected '%s'" % (handler, shortcut)
        with self.assertRaises(InvalidShortcutException, msg=msg):
            self.create_handler(shortcut, handler)

    def test_shortcut_invalid(self):
        for s in INVALID_ALL + VALID_URIS + VALID_PPAS + VALID_UCAS:
            self.check_invalid_shortcut(SourcesListShortcutHandler, s)
        for s in INVALID_ALL + VALID_LINES + VALID_PPAS + VALID_UCAS:
            self.check_invalid_shortcut(URIShortcutHandler, s)
        for s in INVALID_ALL + VALID_LINES + VALID_URIS + VALID_UCAS:
            self.check_invalid_shortcut(PPAShortcutHandler, s)
        for s in INVALID_ALL + VALID_LINES + VALID_URIS + VALID_PPAS:
            self.check_invalid_shortcut(CloudArchiveShortcutHandler, s)
        for s in INVALID_ALL:
            self.check_invalid_shortcut(shortcut_handler, s)


class EnableSourceShortcutsTestcase(ShortcutsTestcase):
    enable_source = True


if __name__ == "__main__":
    unittest.main()
